from dataclasses import dataclass
from pygears.sim.sim_gear import SimGear
from pygears.sim import delta, timestep, log, clk
from pygears.sim.sim import SimPlugin
from pygears.conf import Inject, inject
from pygears import GearDone, reg
from .cosim_port import CosimNoData, InCosimPort, OutCosimPort


class CosimBase(SimGear):
    @inject
    def __init__(self, gear, timeout=-1, sim_map=Inject('sim/map')):
        super().__init__(gear)
        self.timeout = timeout
        self.in_cosim_ports = [InCosimPort(self, p) for p in gear.in_ports]
        self.out_cosim_ports = [OutCosimPort(self, p) for p in gear.out_ports]
        self.eval_needed = False

        for p in (self.in_cosim_ports + self.out_cosim_ports):
            sim_map[p.port] = p

    def cycle(self):
        raise NotImplementedError()

    def forward(self):
        raise NotImplementedError()

    def back(self):
        raise NotImplementedError()

    def read_out(self, port):
        if not port in self.handlers:
            raise ConnectionResetError

        if self.eval_needed:
            self.forward()

        self.eval_needed = True

        hout = self.handlers[port]
        hout.reset()
        return hout.read()

    def ack_out(self, port):
        if not port in self.handlers:
            raise ConnectionResetError

        self.eval_needed = True
        hout = self.handlers[port]
        hout.ack()
        self.activity_monitor = 0

    def write_in(self, port, data):
        if not port in self.handlers:
            raise ConnectionResetError

        self.eval_needed = True

        hin = self.handlers[port]
        return hin.send(data)

    def reset_in(self, port):
        if not port in self.handlers:
            raise ConnectionResetError

        self.eval_needed = True

        hin = self.handlers[port]
        hin.reset()

    def ready_in(self, port):
        if not port in self.handlers:
            raise ConnectionResetError

        if self.eval_needed:
            self.back()
            self.eval_needed = False

        hin = self.handlers[port]
        if hin.ready():
            self.activity_monitor = 0
            return True
        else:
            return False

    async def func(self, *args, **kwds):
        self.activity_monitor = 0
        self.eval_needed = False

        try:
            while True:

                phase = None
                while phase != 'back':
                    phase = await delta()

                if self.eval_needed:
                    self.forward()
                    self.eval_needed = False

                if self.activity_monitor == self.timeout:
                    raise GearDone

                self.cycle()
                self.activity_monitor += 1

        except (GearDone, BrokenPipeError):
            # print(f"SimGear canceling: {self.gear.name}")
            for p in self.gear.out_ports:
                p.producer.finish()

            self._finish()
            raise GearDone


@dataclass
class AuxClock:
    name: str
    frequency: int


class CosimPlugin(SimPlugin):
    @classmethod
    def bind(cls):
        reg.confdef('sim/aux_clock', default=[])



from .base import GenericMeta
from abc import ABCMeta, abstractmethod


class NumberType(ABCMeta, GenericMeta):
    @property
    @abstractmethod
    def signed(self) -> bool:
        ...


class Number(metaclass=NumberType):
    """All numbers inherit from this class.
    If you just want to check if an argument x is a number, without
    caring what kind, use isinstance(x, Number).
    """
    __slots__ = ()

    # Concrete numeric types must provide their own hash implementation
    __hash__ = None


# class Number(int, metaclass=NumberType):
#     pass



from pygears import gear, alternative, find, module
from pygears.sim import clk, delta
from pygears.typing import Unit, cast
from .const import ping


def state_din_setup(module):
    module.val = module.params['init']


@gear(sim_setup=state_din_setup)
async def state_din(din, *, init) -> None:
    async with din as data:
        pass

    await delta()
    module().val = data


def state_dout_setup(module):
    module.state_din = find('../state_din')


@gear(sim_setup=state_dout_setup)
async def state_dout(*rd, t) -> b't':
    dout = [None] * len(rd)
    for i, rd_req in enumerate(rd):
        if not rd_req.empty():
            dout[i] = t(module().state_din.val)

    if len(rd) > 1:
        yield tuple(dout)
    else:
        yield dout[0]

    for i, rd_req in enumerate(rd):
        if dout[i] is not None:
            rd_req.get_nb()

    if all(d is None for d in dout):
        await clk()


@gear(enablement=b'len(rd) > 1', hdl={'hierarchical': False})
def state(din, *rd: Unit, init=0, hold=False) -> b'(din,)*len(rd)':
    din | state_din(init=init)
    return tuple(state_dout(r, t=din.dtype) for r in rd)


@alternative(state)
@gear(hdl={'impl': 'state', 'hierarchical': False})
def state_single_out(din, rd: Unit, *, init=0, hold=False) -> b'din':
    din | state_din(init=init)
    return state_dout(rd, t=din.dtype)


@alternative(state)
@gear
def state_perp(din, *, n, init=0, hold=False):
    return state(din, *(ping(1), ) * n, init=init)



from pygears import reg
from pygears.entry import EntryPlugin, cmd_register


def synth(
    tool,
    design=None,
    outdir=None,
    include=None,
    top='/',
    build=True,
    generate=True,
    **kwds):

    backends = reg['entry/cmds/synth/cmds']
    if tool not in backends:
        raise Exception(f'Unknown backend synth tool "{tool}".')

    return backends[tool]['entry'](
        design=design, include=include, top=top, build=build, outdir=outdir, **kwds)


class SynthPlugin(EntryPlugin):
    @classmethod
    def bind(cls):
        conf = cmd_register(
            ['synth'],
            synth,
            structural=True,
            aliases=['ip'],
            help='synthetize a module',
            dest='tool')

        baseparser = conf['baseparser']

        baseparser.add_argument('design', type=str)
        baseparser.add_argument(
            '-I', '--include', action='append', default=[], help="HDL include directory")
        baseparser.add_argument('--top', '-t', type=str, default='/')
        baseparser.add_argument('--build', '-b', action='store_false')
        baseparser.add_argument('--generate', '-g', action='store_false')
        baseparser.add_argument('--outdir', '-o', type=str)
        baseparser.add_argument(
            '--lang', '-l', type=str, choices=['v', 'sv'], default='sv')



import copy
import functools
import inspect
import pygears

from pygears.conf import Inject, inject, reg, PluginBase
from pygears.core.graph import get_producer_port
from pygears.typing import Tuple, typeof
from pygears import module
from pygears.core.util import is_async_gen

from .funcutils import FunctionMaker
from .partial import Partial
from .util import doublewrap, get_function_context_dict


def add_alternative(base, alter):
    alternatives = getattr(base, 'alternatives', [])
    alternatives.append(alter)
    gear_func = getattr(alter, '__wrapped__', alter)

    if hasattr(gear_func, 'alternatives'):
        alternatives.extend(gear_func.alternatives)

    gear_func_to = getattr(base, '__wrapped__', base)

    gear_func.alternative_to = gear_func_to
    gear_func_to.alternatives = alternatives

    if hasattr(gear_func, 'meta_kwds'):
        gear_func.meta_kwds['__base__'] = gear_func_to

    base.alternatives = alternatives


def alternative(*base_gear_defs):
    def gear_decorator(gear_def):
        for d in base_gear_defs:
            add_alternative(d.func, gear_def.func)
        return gear_def

    return gear_decorator


@inject
def find_invocation(func, extra_params=Inject('gear/params/extra')):
    invocation = []
    sig = inspect.signature(func)

    for name, param in sig.parameters.items():
        if param.kind == param.KEYWORD_ONLY:
            invocation.append(f'{name}={name}')
        elif param.kind == param.VAR_POSITIONAL:
            invocation.append(f'*{name}')
        elif param.kind != param.VAR_KEYWORD:
            invocation.append(name)

    if extra_params:
        for k, v in extra_params.items():
            invocation.append(f'{k}={k}')

    for name, param in sig.parameters.items():
        if param.kind == param.VAR_KEYWORD:
            invocation.append(f'**{name}')

    return ','.join(invocation)


def formatannotation(annotation, base_module=None):
    if getattr(annotation, '__module__', '').startswith('pygears.typing'):
        return repr(annotation)
    if getattr(annotation, '__module__', None) == 'typing':
        return repr(annotation).replace('typing.', '')
    if isinstance(annotation, type):
        if annotation.__module__ in ('builtins', base_module):
            return annotation.__qualname__
        return annotation.__module__ + '.' + annotation.__qualname__
    return repr(annotation)


def formatargspec(args,
                  varargs=None,
                  varkw=None,
                  defaults=None,
                  kwonlyargs=(),
                  kwonlydefaults={},
                  annotations={},
                  formatarg=str,
                  formatvarargs=lambda name: '*' + name,
                  formatvarkw=lambda name: '**' + name,
                  formatvalue=lambda value: '=' + repr(value),
                  formatreturns=lambda text: ' -> ' + text,
                  formatannotation=formatannotation):
    """Format an argument spec from the values returned by getfullargspec.

    The first seven arguments are (args, varargs, varkw, defaults,
    kwonlyargs, kwonlydefaults, annotations).  The other five arguments
    are the corresponding optional formatting functions that are called to
    turn names and values into strings.  The last argument is an optional
    function to format the sequence of arguments.

    Deprecated since Python 3.5: use the `signature` function and `Signature`
    objects.
    """
    def formatargandannotation(arg):
        result = formatarg(arg)
        if arg in annotations:
            result += ': ' + formatannotation(annotations[arg])
        return result

    specs = []
    if defaults:
        firstdefault = len(args) - len(defaults)
    for i, arg in enumerate(args):
        spec = formatargandannotation(arg)
        if defaults and i >= firstdefault:
            spec = spec + formatvalue(defaults[i - firstdefault])
        specs.append(spec)
    if varargs is not None:
        specs.append(formatvarargs(formatargandannotation(varargs)))
    else:
        if kwonlyargs:
            specs.append('*')
    if kwonlyargs:
        for kwonlyarg in kwonlyargs:
            spec = formatargandannotation(kwonlyarg)
            if kwonlydefaults and kwonlyarg in kwonlydefaults:
                spec += formatvalue(kwonlydefaults[kwonlyarg])
            specs.append(spec)
    if varkw is not None:
        specs.append(formatvarkw(formatargandannotation(varkw)))
    result = '(' + ', '.join(specs) + ')'
    if 'return' in annotations:
        result += formatreturns(formatannotation(annotations['return']))
    return result


def create_unpacked_tuple_alternative(g):
    args, *paramspec, annotations = inspect.getfullargspec(g.func)

    if len(args) != 1:
        return

    arg = args[0]

    if arg not in annotations:
        return

    din_type = annotations[arg]

    if not typeof(din_type, Tuple):
        return

    if not din_type.fields:
        return

    unpack_annot = {name: dtype for name, dtype in zip(din_type.fields, din_type.args)}

    signature = formatargspec(din_type.fields, *paramspec)

    f = FunctionMaker(name=f'__{g.func.__name__}_unpack__', signature=signature)

    f.annotations = unpack_annot

    base_func = g.func

    while (hasattr(base_func, 'alternative_to')):
        base_func = base_func.alternative_to

    body = f'''def %(name)s%(signature)s:
    {arg} = ccat({",".join(din_type.fields)})
    try:
        return __{base_func.__name__}({find_invocation(base_func)}, __no_unpack_alt__=True)
    except Exception as e:
        gear_inst = module().child[-1]
        gear_inst.parent.child.remove(gear_inst)
        for port in gear_inst.in_ports:
            if port.basename not in gear_inst.const_args:
                port.producer.consumers.remove(port)
            else:
                gear_inst.parent.child.remove(get_producer_port(port).gear)
        raise e
    '''

    from ..lib.ccat import ccat
    closure = {'ccat': ccat, f'__{base_func.__name__}': g, 'pygears': pygears, 'module': module}
    closure.update(get_function_context_dict(g.func))

    unpack_func = f.make(body, evaldict=closure, addsource=True)

    unpack_func.__kwdefaults__ = paramspec[-1]

    add_alternative(base_func, unpack_func)


def infer_outnames(annotations, meta_kwds):
    outnames = None
    if "return" in annotations:
        if isinstance(annotations['return'], dict):
            outnames = tuple(annotations['return'].keys())

    if not outnames:
        outnames = meta_kwds['outnames']

    if not outnames:
        outnames = []

    return outnames


def create_gear_definition(func, gear_resolver=None, **meta_kwds):

    if inspect.isgeneratorfunction(func):
        raise Exception(f'Generator function {func} cannot be used as a module. PyGears currently only supports regular functions'
                        f' or async generators')

    if gear_resolver is None:
        gear_resolver = reg['gear/gear_dflt_resolver']

    # Add defaults from GearMetaParams registry
    for k, v in reg['gear/params/meta'].items():
        if k not in meta_kwds:
            meta_kwds[k] = copy.copy(v)

    execdict = {'gear_resolver': gear_resolver, 'gear_func': func}
    execdict.update(get_function_context_dict(func))

    invocation = find_invocation(func)
    body = f'return gear_resolver(gear_func, {invocation})'

    gear_func = FunctionMaker.create(
        obj=func,
        body=body,
        evaldict=execdict,
        addsource=True,
        extra_kwds={k: copy.copy(v)
                    for k, v in reg['gear/params/extra'].items()})

    functools.update_wrapper(gear_func, func)

    p = Partial(gear_func)

    meta_kwds['definition'] = p
    meta_kwds['outnames'] = infer_outnames(func.__annotations__, meta_kwds)
    func.meta_kwds = meta_kwds

    create_unpacked_tuple_alternative(p)

    return p


gear = doublewrap(create_gear_definition)


class GearDecoratorPlugin(PluginBase):
    @classmethod
    def bind(cls):
        reg['gear/gear_dflt_resolver'] = None



import weakref

from functools import partial


# def callback(p, b):
#     print(f'Consumer of {p.name} consumer is deleted.')


class Port:
    def __init__(self, gear, index, basename, dtype=None, producer=None, consumer=None):
        self.gear = gear
        self.index = index
        self.producer = producer
        self._dtype = dtype
        # self._consumer = None if consumer is None else weakref.ref(consumer, partial(callback, self))
        self._consumer = None if consumer is None else weakref.ref(consumer)
        self.basename = basename

    @property
    def consumer(self):
        return self._consumer if self._consumer is None else self._consumer()

    @consumer.setter
    def consumer(self, val):
        # self._consumer = weakref.ref(val, partial(callback, self))
        self._consumer = weakref.ref(val)

    @property
    def name(self):
        return f'{self.gear.name}.{self.basename}'

    @property
    def dtype(self):
        if self._dtype is not None:
            return self._dtype

        # TODO: Remove this later
        if self.producer is not None:
            return self.producer.dtype
        else:
            return self.consumer.dtype

    def finish(self):
        if self.consumer:
            self.consumer.finish()


class InPort(Port):
    direction = "in"

    @property
    def dtype(self):
        if self._dtype is not None:
            return self._dtype

        if self.consumer is not None:
            return self.consumer.dtype
        else:
            return self.producer.dtype

    def __repr__(self):
        return f'InPort("{self.name}")'


class OutPort(Port):
    direction = "out"

    def __repr__(self):
        return f'OutPort("{self.name}")'


class HDLUser:
    pass


class HDLProducer(HDLUser):
    pass


class HDLConsumer(HDLUser):
    pass



import runpy
import re
import os
import shutil
from pygears.hdl import hdlgen, list_hdl_files
from pygears import reg, find
from pygears.util.fileio import get_main_script
from pygears.hdl.synth import SynthPlugin
from pygears.entry import cmd_register
from pygears.conf.custom_settings import load_rc
from pygears.util.fileio import find_in_dirs

black_box_modules = ['tdp', 'sdp', 'decouple', 'fifo']

def get_file_name(f, outdir, lang):
    path = find_in_dirs(f, reg[f'{lang}gen/include'])
    if path is not None:
        return path
    else:
        return os.path.join(outdir, f)

def get_top_file_name(node, outdir):
    return get_file_name(node.wrap_file_name, outdir, node.lang)

def create_project_script(script_fn, outdir, top):

    black_box = set()
    sub_black_box = set()

    def black_box_filt(node):
        for name in black_box_modules:
            bb = False
            if name[0] == '/':
                if node.node.name.startswith(name):
                    bb = True
            else:
                if node.module_name == name:
                    bb = True

            if bb:
                black_box.add(get_top_file_name(node, outdir))
                for f in node.files:
                    sub_black_box.add(get_file_name(f, outdir, node.lang))

                return False

        return True

    hdl_files = list_hdl_files(top, outdir, filt=black_box_filt)

    with open(script_fn, 'w') as f:
        for fn in hdl_files:
            f.write(f'read_verilog -sv {os.path.abspath(fn)}\n')

        for fn in black_box:
            print(f"Loading {fn} as blackbox")
            f.write(f'read_verilog -sv -lib {fn}\n')

    return list(sub_black_box)


class Yosys:
    PROMPT = 'yosys>'
    RE_INIT_ATTR = re.compile(r"attribute \\init (\d+)'(\d+)")
    RE_REG_DEF = re.compile(r"wire width (\d+) (\S+)")

    RE_STATS_LUT = re.compile(r"LUT(\d)\s+(\d+)")
    RE_STATS_FDRE = re.compile(r"FDRE\s+(\d+)")
    RE_STATS_DRAM = re.compile(r"(?:RAM64X1D|RAM32X1D)\s+(\d+)")

    def __init__(self, cmd_line):
        self.cmd_line = cmd_line

    def __enter__(self):
        import pexpect
        self.proc = pexpect.spawnu(self.cmd_line)
        self.proc.expect(Yosys.PROMPT)
        self.proc.setecho(False)
        self.EOF = pexpect.EOF
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.proc.close()

    def command(self, cmd):
        self.proc.sendline(cmd)
        try:
            self.proc.expect(Yosys.PROMPT, timeout=None)
        except self.EOF:
            print(self.proc.before)
            raise

        # print(self.proc.before.strip())
        return self.proc.before.strip()

    def enum_registers(self):
        current_init_value = None
        regs = {}
        ret = self.command('dump t:$dff %x:+[Q] t:$dff %d')
        for line in ret.strip().split('\n'):
            line = line.strip()

            ret = Yosys.RE_INIT_ATTR.search(line)
            if ret:
                current_init_value = ret.groups(0)[1]
                continue

            ret = Yosys.RE_REG_DEF.search(line)
            if ret:
                width, name = ret.groups(0)
                regs[name] = (width, current_init_value)
                continue

        return regs

    @property
    def stats(self):
        res = {'logic luts': 0, 'ffs': 0, 'lutrams': 0}

        ret = self.command('stat')
        for line in ret.strip().split('\n'):
            ret = Yosys.RE_STATS_LUT.search(line)
            if ret:
                res['logic luts'] += int(ret.groups(0)[1])
                continue

            ret = Yosys.RE_STATS_FDRE.search(line)
            if ret:
                res['ffs'] += int(ret.groups(0)[0])
                continue

            ret = Yosys.RE_STATS_DRAM.search(line)
            if ret:
                res['lutrams'] += int(ret.groups(0)[0])
                continue

        return res


def synth(outdir,
          srcdir=None,
          top=None,
          optimize=True,
          freduce=False,
          synthout=None,
          blackbox=None,
          lang='v',
          synthcmd='synth'):

    if blackbox:
        # TODO: Don't go global with this
        black_box_modules.extend(blackbox)

    if not srcdir:
        srcdir = outdir

    prj_script_fn = os.path.join(outdir, 'project.ys')

    # synth_out_fn = os.path.join(outdir, 'synth.v')

    if isinstance(top, str):
        top_mod = find(top)
    else:
        top_mod = top

    vgen_map = reg['hdlgen/map']

    if top_mod not in vgen_map:
        hdlgen(top=top_mod, lang=lang, toplang='v', outdir=srcdir)

    vgen_inst = vgen_map[top_mod]

    if vgen_inst.parent_lang == 'v':
        top_name = vgen_inst.wrap_module_name
    else:
        top_name = vgen_inst.inst_name

    black_box = create_project_script(prj_script_fn,
                                      outdir=srcdir,
                                      top=top_mod)
    if synthout:
        for b in black_box:
            shutil.copy(b, outdir)

    with Yosys(f'yosys -l {os.path.join(outdir, "yosys.log")}') as yosys:

        yosys.command(f'script {prj_script_fn}')

        yosys.command(f'hierarchy -check -top {top_name}')

        yosys.command(f'proc')
        yosys.command(f'flatten')

        if optimize:
            yosys.command(f'opt -full -sat')

            if freduce:
                yosys.command(f'freduce')
                yosys.command(f'opt -full')

        if synthcmd:
            yosys.command(synthcmd)

        if synthout:
            yosys.command(f'clean -purge')
            yosys.command(f'write_verilog -noattr {synthout}')

        return yosys.stats


def entry(
        top,
        design,
        outdir=None,
        srcdir=None,
        include=None,
        lang='sv',
        generate=True,
        freduce=False,
        synthout=None,
        synthcmd='synth',
        blackbox='',
        build=True):

    if reg['yosys/synth/lock']:
        return

    if design is None:
        design = get_main_script()

    if design is not None:
        design = os.path.abspath(os.path.expanduser(design))

    # TODO: Outdir can be unspecified, use global results-dir then
    os.makedirs(outdir, exist_ok=True)

    # makefile(
    #     top,
    #     design,
    #     outdir,
    #     lang=lang,
    #     generate=generate,
    #     build=build,
    #     include=include,
    #     prjdir=prjdir)

    if not generate:
        return

    if isinstance(top, str):
        top_mod = find(top)
    else:
        top_mod = top

    if top_mod is None:
        reg['yosys/synth/lock'] = True
        load_rc('.pygears', os.path.dirname(design))
        runpy.run_path(design)
        reg['yosys/synth/lock'] = False
        top_mod = find(top)

    if top_mod is None:
        raise Exception(
            f'Module "{top}" specified as a IP core top level module, not found in the design "{design}"')

    if include is None:
        include = []

    include += reg[f'{lang}gen/include']

    if blackbox:
        blackbox = blackbox.split(',')
    else:
        blackbox = None

    report = synth(outdir, top=top_mod, lang=lang, synthout=synthout, synthcmd=synthcmd, freduce=freduce, srcdir=srcdir, blackbox=blackbox)

    return report


class YosysSynthPlugin(SynthPlugin):
    @classmethod
    def bind(cls):
        conf = cmd_register(['synth', 'yosys'], entry, derived=True)

        reg['yosys/synth/lock'] = False

        conf['parser'].add_argument('--synthout', type=str)
        conf['parser'].add_argument('--srcdir', type=str)
        conf['parser'].add_argument('--synthcmd', type=str)
        conf['parser'].add_argument('--blackbox', type=str)
        # conf['parser'].add_argument('--util', action='store_true')
        # conf['parser'].add_argument('--timing', action='store_true')



from math import floor, ceil
from .base import class_and_instance_method, typeof, is_type
from .unit import Unit
from .uint import IntegralType, Integral, Uint, Int, Integer, code, Bool
from .math import bitw


class FixpnumberType(IntegralType):
    def __abs__(self):
        if not self.signed:
            return self

        return Fixp[self.integer + 1, self.width + 1]

    def __add__(self, other):
        if not typeof(other, Integral):
            return NotImplemented

        signed = self.signed or other.signed
        res_type = Fixp if signed else Ufixp

        i1 = self.integer + 1 if signed and not self.signed else self.integer
        fr1 = self.fract

        if typeof(other, Fixpnumber):
            i2 = other.integer + 1 if signed and not other.signed else other.integer
            fr2 = other.fract
        elif typeof(other, Integer):
            i2 = other.width + 1 if signed and not other.signed else other.width
            fr2 = 0
        else:
            return NotImplemented

        integer_part = max(i1, i2) + 1
        fract_part = max(fr1, fr2)
        width = integer_part + fract_part

        return res_type[integer_part, width]

    def __and__(self, other):
        return NotImplemented

    def __ceil__(self):
        if self.fract > 0:
            return self.base[self.integer + 1, self.width + 1]
        else:
            return self

    def __float__(self):
        return float

    def __floordiv__(self, other):
        return self.__truediv__(other, 0)

    def __floor__(self):
        return self

    def __ge__(self, other):
        return Bool

    def __getitem__(self, index):
        if not self.specified:
            return super().__getitem__(index)

        index = self.index_norm(index)

        width = 0
        for i in index:
            if isinstance(i, slice):
                if i.stop == 0:
                    return Unit
                elif i.stop - i.start > len(self):
                    raise IndexError
                width += i.stop - i.start
            else:
                if i >= len(self):
                    raise IndexError
                width += 1

        return Uint[width]

    def __gt__(self, other):
        return Bool

    def __iadd__(self, other):
        return self

    def __int__(self):
        return int

    def __invert__(self):
        return NotImplemented

    def __imul__(self, other):
        return self

    def __isub__(self, other):
        return self

    def __lshift__(self, others):
        shamt = int(others)
        return self.base[self.integer + shamt, self.width]

    def __le__(self, other):
        return Bool

    def __lt__(self, other):
        return Bool

    def __mod__(self, other):
        return NotImplemented

    def __mul__(self, other):
        if not typeof(other, Integral):
            return NotImplemented

        if typeof(other, Integer):
            other_cls = Fixp if other.signed else Ufixp
            other = other_cls[other.width, other.width]

        ops = [self, other]

        signed = any(op.signed for op in ops)

        try:
            integer_part = self.integer + other.integer
            fract_part = self.fract + other.fract
        except AttributeError:
            integer_part = self.integer + int(other)
            fract_part = self.fract

        width = integer_part + fract_part

        if signed:
            return Fixp[integer_part, width]
        else:
            return Ufixp[integer_part, width]

    def __neg__(self):
        return Fixp[self.integer + 1, self.width + 1]

    def __or__(self, other):
        return NotImplemented

    __radd__ = __add__

    def __rand__(self, other):
        return NotImplemented

    def __rfloordiv__(self, other):
        return NotImplemented

    def __rlshift__(self, other):
        return NotImplemented

    def __rmod__(self, other):
        return NotImplemented

    __rmul__ = __mul__

    def __ror__(self, other):
        return NotImplemented

    def __round__(self, digits=0):
        if self.fract > digits:
            return self.base[self.integer + 1, self.width + 1]
        else:
            return self

    def __rshift__(self, others):
        shamt = int(others)
        return self.base[self.integer - shamt, self.width]

    def __rrshift__(self, other):
        return NotImplemented

    def __rtruediv__(self, other):
        return NotImplemented

    def __rxor__(self, other):
        return NotImplemented

    def __str__(self):
        if self.args:
            return f'q{self.integer}.{self.width - self.integer}'
        else:
            return super().__str__()

    def __sub__(self, other):
        if not typeof(other, Integral):
            return NotImplemented

        i1 = self.integer + 1 if other.signed and not self.signed else self.integer
        fr1 = self.fract

        if typeof(other, Fixpnumber):
            i2 = other.integer
            fr2 = other.fract
        elif typeof(other, Integer):
            i2 = other.width
            fr2 = 0
        else:
            return NotImplemented

        integer_part = max(i1, i2) + 1
        fract_part = max(fr1, fr2)
        width = integer_part + fract_part

        return Fixp[integer_part, width]

    def __truediv__(self, other, subprec=0):
        ops = [self, other]

        signed = any(op.signed for op in ops)

        try:
            integer_part = self.integer + other.fract
        except AttributeError:
            integer_part = self.integer

        width = self.width + subprec

        if signed:
            return Fixp[integer_part, width]
        else:
            return Ufixp[integer_part, width]

    def __trunc__(self, other):
        if self.signed:
            return Int[self.integer]
        else:
            return Uint[self.integer]

    def keys(self):
        """Returns a list of keys that can be used for indexing the type.

        >>> Int[8].keys()
        [0, 1, 2, 3, 4, 5, 6, 7]
        """
        return list(range(self.width))

    @property
    def integer(self) -> int:
        return self.__args__[0]

    @property
    def fract(self) -> int:
        return self.width - self.integer

    @property
    def specified(self):
        return False

    @property
    def width(self) -> int:
        return self.__args__[1]


class Fixpnumber(Integral, metaclass=FixpnumberType):
    def __new__(cls, val=0):
        # TODO: Instantiating a type that has templated arguments unresolved
        # should throw a nicer error

        # TODO: this takes some time. Implement hashing for classes to speed things up
        if type(val) is cls:
            return val

        if cls.is_generic():
            if cls.is_abstract():
                if isinstance(val, Integer):
                    cls = Fixp if val.signed else Ufixp
                else:
                    try:
                        val = int(val)
                        cls = Fixp
                    except:
                        raise TypeError(f'Unsupported value {val} of type {type(val)}')

            if not is_type(type(val)):
                return cls[bitw(val) + 1, bitw(val) + 1](int(val))
            elif isinstance(val, Integer):
                return cls[val.width, val.width].decode(val.code())

        if isinstance(val, Fixpnumber):
            val_fract = type(val).fract
            if cls.fract > val_fract:
                ival = int(val) << (cls.fract - val_fract)
            else:
                ival = int(val) >> (val_fract - cls.fract)
        elif isinstance(val, Integer):
            ival = round(float(val) * (2**cls.fract))
        elif isinstance(val, float):
            ival = round(float(val) * (2**cls.fract))
        else:
            try:
                ival = round(float(val) * (2**cls.fract))
            except TypeError:
                raise TypeError(f'Unsupported value {val} of type {type(val)}')

        if cls.signed:
            if (bitw(ival) > (cls.width if ival < 0 else cls.width - 1)):
                raise ValueError(f"{repr(cls)} cannot represent value '{val}'")
        else:
            if ival < 0:
                raise ValueError(
                    f"cannot represent negative numbers with unsigned type '{repr(cls)}'")

            if (bitw(ival) > cls.width):
                raise ValueError(f"{repr(cls)} cannot represent value '{val}'")

        res = int.__new__(cls, ival)

        return res

    def __abs__(self):
        if self.signed:
            return -self if code(self, Uint)[-1] else self
        else:
            return self

    def __add__(self, other):
        if is_type(type(other)) and not isinstance(other, (Fixpnumber, Integer)):
            return NotImplemented

        if isinstance(other, Fixpnumber):
            conv_other = other
        else:
            conv_other = Fixpnumber(other)

        sum_cls = type(self) + type(conv_other)
        return sum_cls.decode((int(self) << (sum_cls.fract - type(self).fract)) +
                              (int(conv_other) << (sum_cls.fract - type(conv_other).fract)))

    def __and__(self, other):
        return NotImplemented

    def __ceil__(self):
        t = type(self)
        if t.fract >= 1:
            i = int(self)
            fract = i & ((1 << t.fract) - 1)
            if fract:
                return ceil(t).decode(((i >> (t.fract)) + 1) << t.fract)
            else:
                return ceil(t).decode(i)
        else:
            return self

    def __eq__(self, other):
        if isinstance(other, float):
            return float(self) == other
        elif isinstance(other, Fixpnumber):
            if type(other).fract > type(self).fract:
                return int(self) << (type(other).fract - type(self).fract) == int(other)
            else:
                return int(self) == int(other) << (type(self).fract - type(other).fract)
        else:
            fixp_other = Fixpnumber(other)

            if type(fixp_other).fract > type(self).fract:
                return int(self) << (type(fixp_other).fract - type(self).fract) == int(fixp_other)
            else:
                return int(self) == int(fixp_other) << (type(self).fract - type(fixp_other).fract)

    def __float__(self):
        return int(self) / (2**type(self).fract)

    def __floor__(self):
        t = type(self)
        if t.fract >= 1:
            return floor(t).decode((int(self) >> t.fract) << t.fract)
        else:
            return self

    def __floordiv__(self, other):
        return self.__truediv__(other, 0)

    def __ge__(self, other):
        if isinstance(other, float):
            return float(self) >= other
        elif isinstance(other, Fixpnumber):
            if type(other).fract > type(self).fract:
                return int(self) << (type(other).fract - type(self).fract) >= int(other)
            else:
                return int(self) >= int(other) << (type(self).fract - type(other).fract)
        else:
            fixp_other = Fixpnumber(other)

            if type(fixp_other).fract > type(self).fract:
                return int(self) << (type(fixp_other).fract - type(self).fract) >= int(fixp_other)
            else:
                return int(self) >= int(fixp_other) << (type(self).fract - type(fixp_other).fract)

    def __gt__(self, other):
        if isinstance(other, float):
            return float(self) > other
        elif isinstance(other, Fixpnumber):
            if type(other).fract > type(self).fract:
                return int(self) << (type(other).fract - type(self).fract) > int(other)
            else:
                return int(self) > int(other) << (type(self).fract - type(other).fract)
        else:
            fixp_other = Fixpnumber(other)

            if type(fixp_other).fract > type(self).fract:
                return int(self) << (type(fixp_other).fract - type(self).fract) > int(fixp_other)
            else:
                return int(self) > int(fixp_other) << (type(self).fract - type(fixp_other).fract)

    def __iadd__(self, other):
        if is_type(type(other)) and not isinstance(other, (Fixpnumber, Integer)):
            return NotImplemented

        if isinstance(other, Fixpnumber):
            conv_other = other
        else:
            conv_other = Fixpnumber(other)

        sum_cls = type(self)
        return sum_cls.decode((int(self) << (sum_cls.fract - type(self).fract)) +
                              (int(conv_other) << (sum_cls.fract - type(conv_other).fract)))


    def __imul__(self, other):
        if is_type(type(other)) and not isinstance(other, (Fixpnumber, Integer)):
            return NotImplemented

        if isinstance(other, Fixpnumber):
            conv_other = other
        else:
            conv_other = Fixpnumber(other)

        return type(self).decode(super().__mul__(conv_other))

    def __invert__(self):
        return NotImplemented

    def __isub__(self, other):
        if is_type(type(other)) and not isinstance(other, (Fixpnumber, Integer)):
            return NotImplemented

        if isinstance(other, Fixpnumber):
            conv_other = other
        else:
            conv_other = Fixpnumber(other)

        sum_cls = type(self)
        return sum_cls.decode((int(self) << (sum_cls.fract - type(self).fract)) -
                              (int(conv_other) << (sum_cls.fract - type(conv_other).fract)))

    def __hash__(self):
        return hash((type(self), int(self)))

    def __le__(self, other):
        if isinstance(other, float):
            return float(self) <= other
        elif isinstance(other, Fixpnumber):
            if type(other).fract > type(self).fract:
                return int(self) << (type(other).fract - type(self).fract) < int(other)
            else:
                return int(self) <= int(other) << (type(self).fract - type(other).fract)
        else:
            fixp_other = Fixpnumber(other)

            if type(fixp_other).fract > type(self).fract:
                return int(self) << (type(fixp_other).fract - type(self).fract) < int(fixp_other)
            else:
                return int(self) <= int(fixp_other) << (type(self).fract - type(fixp_other).fract)

    def __lshift__(self, other):
        return (type(self) << other).decode(int(self))

    # TODO: What about the comparison with float
    def __lt__(self, other):
        if isinstance(other, float):
            return float(self) < other
        elif isinstance(other, Fixpnumber):
            if type(other).fract > type(self).fract:
                return int(self) << (type(other).fract - type(self).fract) < int(other)
            else:
                return int(self) < int(other) << (type(self).fract - type(other).fract)
        else:
            fixp_other = Fixpnumber(other)

            if type(fixp_other).fract > type(self).fract:
                return int(self) << (type(fixp_other).fract - type(self).fract) < int(fixp_other)
            else:
                return int(self) < int(fixp_other) << (type(self).fract - type(fixp_other).fract)

    def __mod__(self, other):
        return NotImplemented

    def __mul__(self, other):
        if is_type(type(other)) and not isinstance(other, (Fixpnumber, Integer)):
            return NotImplemented

        if isinstance(other, Fixpnumber):
            conv_other = other
        else:
            conv_other = Fixpnumber(other)

        return (type(self) * type(conv_other)).decode(super().__mul__(conv_other))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __neg__(self):
        return (-type(self)).decode(super().__neg__())

    def __or__(self, other):
        return NotImplemented

    def __rand__(self, other):
        return NotImplemented

    __radd__ = __add__

    def __repr__(self):
        return f'{repr(type(self))}({float(self)})'

    def __rfloordiv__(self, other):
        return NotImplemented

    def __rlshift__(self, other):
        return NotImplemented

    def __rmod__(self, other):
        return NotImplemented

    __rmul__ = __mul__

    # TODO: implement digits functionality
    def __round__(self, digits=0):
        t = type(self)
        if t.fract >= 1:
            return round(t).decode((((int(self) >> (t.fract - 1)) + 1) >> 1) << t.fract)
        else:
            return self

    def __rshift__(self, other):
        return (type(self) >> other).decode(int(self))

    def __rsub__(self, other):
        if is_type(type(other)) and not isinstance(other, (Fixpnumber, Integer)):
            return NotImplemented

        if isinstance(other, Fixpnumber):
            conv_other = other
        else:
            conv_other = Fixpnumber(other)

        sum_cls = type(conv_other) - type(self)

        return sum_cls.decode((int(conv_other) << (sum_cls.fract - type(conv_other).fract)) -
                              (int(self) << (sum_cls.fract - type(self).fract)))

    def __str__(self):
        return f'{str(type(self))}({float(self)})'

    def __sub__(self, other):
        if is_type(type(other)) and not isinstance(other, (Fixpnumber, Integer)):
            return NotImplemented

        if isinstance(other, Fixpnumber):
            conv_other = other
        else:
            conv_other = Fixpnumber(other)

        sum_cls = type(self) - type(conv_other)
        return sum_cls.decode((int(self) << (sum_cls.fract - type(self).fract)) -
                              (int(conv_other) << (sum_cls.fract - type(conv_other).fract)))

    @class_and_instance_method
    def __truediv__(self, other, subprec=0):
        if not isinstance(other, Fixpnumber):
            other = Fixpnumber(other)

        div_cls = type(self).__truediv__(type(other), subprec)
        shift = div_cls.fract - type(self).fract + type(other).fract
        return div_cls.decode((self.code() << shift) // other.code())

    def code(self):
        return int(self) & ((1 << self.width) - 1)

    @classmethod
    def decode(cls, val):
        return int.__new__(cls, int(val) & ((1 << cls.width) - 1))

    @property
    def width(self):
        return type(self).width


class FixpType(FixpnumberType):
    @property
    def signed(self):
        return True

    def is_abstract(self):
        return False

    @property
    def specified(self):
        return IntegralType.specified.fget(self)

    @property
    def max(self):
        return self.decode(2**(self.width - 1) - 1)

    @property
    def min(self):
        return self.decode(-2**(self.width - 1))

    @property
    def quant(self):
        return self.decode(1)

    @property
    def fmax(self):
        return (2**(self.width - 1) - 1) / (2**self.fract)

    @property
    def fmin(self):
        return (-2**(self.width - 1)) / (2**self.fract)


class Fixp(Fixpnumber, metaclass=FixpType):
    __parameters__ = ['I', 'W']

    @class_and_instance_method
    @property
    def signed(self):
        return True

    @classmethod
    def decode(cls, val):
        val = int(val)
        if val >= (1 << (cls.width - 1)):
            val -= 1 << cls.width

        return int.__new__(cls, val)


class UfixpType(FixpnumberType):
    @property
    def signed(self) -> bool:
        return False

    def is_abstract(self):
        return False

    @property
    def specified(self) -> bool:
        return IntegralType.specified.fget(self)

    @property
    def max(self):
        return self.decode(2**self.width - 1)

    @property
    def min(self):
        return self.decode(0)

    @property
    def quant(self):
        return self.decode(1)

    @property
    def fmax(self):
        return (2**self.width - 1) / (2**self.fract)

    @property
    def fmin(self):
        return float(0)

    def decode(self, val):
        return int.__new__(self, int(val) & ((1 << self.width) - 1))


class Ufixp(Fixpnumber, metaclass=UfixpType):
    __parameters__ = ['I', 'W']

    @class_and_instance_method
    @property
    def signed(self) -> bool:
        return False



import sys
import inspect
import atexit
from pygears.conf.trace import register_exit_hook, TraceException, make_traceback, TraceLevel
from pygears import reg, GearDone, Intf
from pygears.core.port import HDLConsumer
from pygears.core.channel import report_out_dangling
from pygears.sim import clk, timestep
from pygears.sim.sim import schedule_to_finish
from pygears.conf.trace import gear_definition_location


def is_async_gen(func):
    return bool(func.__code__.co_flags & inspect.CO_ASYNC_GENERATOR)


def is_simgear_func(func):
    return (
        inspect.isgeneratorfunction(func) or inspect.iscoroutinefunction(func)
        or is_async_gen(func))


class SimulationError(TraceException):
    pass


class SimGear:
    def __init__(self, gear):
        self.gear = gear
        self.done = False
        self.child = []
        self.parent = None
        self._clean = True
        self.async_gen = None
        if not hasattr(self, 'func'):
            self.func = gear.func

        for p in gear.out_ports:
            if p.consumer.consumers:
                continue

            report_out_dangling(p)

    def setup(self):
        self._clean = False
        atexit.register(self._finish)
        register_exit_hook(self._cleanup)
        if self.gear.meta_kwds['sim_setup'] is not None:
            self.gear.meta_kwds['sim_setup'](self.gear)

        self.args, self.kwds = self.sim_func_args

    async def run(self):
        # TODO: Try also to detect endless loops
        from pygears.hls.ast.utils import is_func_empty
        if is_func_empty(self.gear.func):
            raise NotImplementedError(f'Cannot run unimplemented gear "{self.gear.name}"')

        out_prods = [p.producer for p in self.gear.out_ports]
        single_output = len(out_prods) == 1
        if single_output:
            out_prods = out_prods[0]

        sim = reg['sim/simulator']
        err = None
        trace_debug = reg['trace/level'] == TraceLevel.debug

        try:
            if is_async_gen(self.func):
                while (1):
                    if sim.phase != 'forward':
                        await clk()

                    self.async_gen = self.func(*self.args, **self.kwds)

                    async for val in self.async_gen:
                        if sim.phase != 'forward':
                            await clk()

                        # TODO: Yielding an interface might make sense to
                        # connect internaly instantiated module directly to the
                        # output, think about it
                        if val is not None:
                            if single_output:
                                tb = None
                                if trace_debug:
                                    out_prods.put_nb(val)
                                else:
                                    try:
                                        out_prods.put_nb(val)
                                    except GearDone:
                                        raise
                                    except Exception as e:
                                        func, fn, ln, _ = gear_definition_location(self.func)

                                        err = SimulationError(
                                            f"inside '{self.gear.name}': {repr(e)}",
                                            self.async_gen.ag_frame.f_lineno,
                                            filename=fn)

                                        traceback = make_traceback((SimulationError, err, sys.exc_info()[2]))
                                        exc_type, exc_value, tb = traceback.standard_exc_info

                                    if tb is not None:
                                        raise exc_value.with_traceback(tb)

                                await out_prods.ready()
                            else:
                                for p, v in zip(out_prods, val):
                                    if v is not None:
                                        p.put_nb(v)

                                for p, v in zip(out_prods, val):
                                    if v is not None:
                                        await p.ready()

                    if self.parent:
                        raise GearDone

                    if self.args:
                        if all(a.done for a in self.args):
                            raise GearDone
            elif inspect.isgeneratorfunction(self.func):
                while (1):
                    if sim.phase != 'forward':
                        await clk()

                    for val in self.func(*self.args, **self.kwds):
                        if sim.phase != 'forward':
                            await clk()

                        if val is not None:
                            if single_output:
                                out_prods.put_nb(val)
                                await out_prods.ready()
                            else:
                                for p, v in zip(out_prods, val):
                                    if v is not None:
                                        p.put_nb(v)

                                for p, v in zip(out_prods, val):
                                    if v is not None:
                                        await p.ready()

                    if self.parent:
                        raise GearDone

                    if self.args:
                        if all(a.done for a in self.args):
                            raise GearDone
            else:
                while (1):
                    # TODO: handle case where self.func is empty and this loop
                    # will just go on forever freezing everything. It happens
                    # if the user accidently creates empty "async def" function
                    await self.func(*self.args, **self.kwds)

                    if self.args:
                        if all(a.done for a in self.args):
                            raise GearDone

        except GearDone as e:
            for p in self.gear.in_ports:
                intf = p.consumer
                try:
                    if intf.producer.producer and not intf.empty():
                        prod_intf = intf.in_queue.intf
                        prod_gear = prod_intf.consumers[0].gear
                        schedule_to_finish(prod_gear)
                except GearDone:
                    pass

            if self.parent:
                self.parent.child.remove(self)

            self._finish()
            raise e
        except Exception as e:
            e.args = (f'{str(e)}, in the module "{self.gear.name}"', )
            err = e

        if err:
            raise err

    @property
    def sim_func_args(self):
        args = []
        for p in self.gear.in_ports:
            i = p.consumer
            if i is None:
                i = Intf(p.dtype)
                i.source(p)
                i.connect(HDLConsumer())

            args.append(i)

        return args, self.gear.explicit_params

    def _finish(self):
        self.done = True
        try:
            for port in self.gear.out_ports:
                port.producer.finish()
        except:
            pass

        if not self._clean:
            self._cleanup()

        atexit.unregister(self._finish)

    def _cleanup(self):
        self._clean = True



import typing
from . import Context, GearContext, FuncContext, Function, Submodule, ir, ir_utils, node_visitor, visit_ast, visit_block
from ..debug import print_func_parse_intro
from pygears import Intf, reg
from pygears.typing import typeof, Bool, Uint
from pygears.core.partial import combine_arg_kwds, extract_arg_kwds
from pygears.core.port import HDLConsumer, HDLProducer


def form_gear_args(args, kwds, func):
    kwd_args, kwds_only = extract_arg_kwds(kwds, func)
    args_only = combine_arg_kwds(args, kwd_args, func)

    return args_only, kwds_only


class UsedVarVisitor(ir_utils.IrExprVisitor):
    def __init__(self, used):
        self.used = used

    def visit_Name(self, node):
        if node.ctx == 'load':
            self.used.add(node.name)


class UsedVarStmtVisitor(ir_utils.IrVisitor):
    def __init__(self):
        super().__init__()
        self.used = set()

    def Expr(self, node):
        UsedVarVisitor(self.used).visit(node)


# TODO: Make it work with multiple target assignements
class UnusedVarCleanup(ir_utils.IrRewriter):
    def __init__(self, used, ctx):
        super().__init__()
        self.used = used
        self.ctx = ctx

    def AssignValue(self, node):
        if not isinstance(node.target, ir.Name):
            return node

        if node.target.name == '_state':
            # TODO: Find a better way of detecting whether state register is not needed
            # If we are getting rid of state register
            if '_rst_cond' in self.ctx.scope:
                state_en = ir.AssignValue(self.ctx.ref('_state_en'), ir.res_true)
                rst_cond = ir.AssignValue(ir.SubscriptExpr(self.ctx.ref('_rst_cond'), node.val),
                                          ir.res_true)
                return [state_en, rst_cond]
            else:
                return node

        if node.target.name in self.used or (isinstance(self.ctx, GearContext)
                                             and node.target.name in self.ctx.intfs):
            return node

        if node.target.name in self.ctx.scope:
            del self.ctx.scope[node.target.name]

        return None


def removed_unused_vars(node, ctx):
    # TODO: This part is only for the cleanup of the whole module with inferred
    # states. Find a better way to implement this if statement
    if isinstance(ctx, GearContext):
        used = set()
        for state in node.stmts[0].branches:
            v = UsedVarStmtVisitor()
            for stmt in state.stmts:
                v.visit(stmt)

            used.update(v.used)

        if len(node.stmts[0].branches) == 1 and '_state' not in used:
            node.stmts = node.stmts[0].branches[0].stmts

            ctx.scope['_state_en'] = ir.Variable('_state_en', dtype=Bool)
            state_type = ctx.scope['_state'].dtype
            state_num = int(state_type.max) + 1
            ctx.scope['_rst_cond'] = ir.Variable('_rst_cond', dtype=Uint[state_num])
            node.stmts.insert(0, ir.AssignValue(ctx.ref('_state_en'), ir.res_false))
            del ctx.scope['_state']

            # TODO: Why can't we just do this? Default value extraction in
            # svcompile creates a problem. Run "test_double_loop"
            # node.stmts.insert(0, ir.AssignValue(ctx.ref('_rst_cond'), ir.ResExpr(ctx.scope['_rst_cond'].dtype(0))))

            for s in range(state_num):
                node.stmts.insert(
                    0,
                    ir.AssignValue(ir.SubscriptExpr(
                        ctx.ref('_rst_cond'), ir.ResExpr(state_type(s))),
                        ir.res_false))

            used.update(['_state_en', '_rst_cond'])

        return UnusedVarCleanup(used, ctx).visit(node)
    else:
        v = UsedVarStmtVisitor()
        v.visit(node)
        return UnusedVarCleanup(v.used, ctx).visit(node)


class ComplexityExplorer(ir_utils.IrExprVisitor):
    def __init__(self):
        self.operations = 0
        self.names = 0

    def visit_Name(self, node):
        self.names += 1

    def visit_BinOpExpr(self, node):
        self.operations += 1
        super().visit_BinOpExpr(node)

    def visit_ConditionalExpr(self, node):
        self.operations += 1
        super().visit_ConditionalExpr(node)


# TODO: Should we use Inliner from passes/inline_cfg?
class Inliner(ir_utils.IrExprRewriter):
    def __init__(self, forwarded):
        self.forwarded = forwarded

    def visit_Name(self, node):
        if ((node.name not in self.forwarded) or (node.ctx != 'load')):
            return None

        val = self.forwarded[node.name]

        if isinstance(val, ir.ResExpr) and getattr(val.val, 'unknown', False):
            return node

        return val


def should_inline(func_ir, func_ctx, args):
    if len(func_ir.stmts) > 1:
        return False

    s = func_ir.stmts[0]

    if not isinstance(s, ir.FuncReturn):
        return False

    v = ComplexityExplorer()
    v.visit(s.expr)

    if v.operations <= 2 and v.names <= len(args):
        return True

    return False


def inline_expr(func_ir, func_ctx, args):
    s = func_ir.stmts[-1]
    return Inliner(args).visit(s.expr)


def parse_func_call(func: typing.Callable, args, kwds, ctx: Context):
    ctx_stack = reg['hls/ctx']

    uniqueid = ''
    if isinstance(ctx, FuncContext):
        uniqueid = ctx.funcref.name

    uniqueid += str(len(ctx_stack[0].functions))

    funcref = Function(func, args, kwds, uniqueid=uniqueid)

    if funcref not in ctx_stack[0].functions:
        func_ctx = FuncContext(funcref, args, kwds)
        print_func_parse_intro(func, funcref.ast)
        reg['hls/ctx'].append(func_ctx)
        func_ir = visit_ast(funcref.ast, func_ctx)
        reg['hls/ctx'].pop()
        func_ir = removed_unused_vars(func_ir, func_ctx)
        ctx_stack[0].functions[funcref] = (func_ir, func_ctx)
    else:
        (func_ir, func_ctx) = ctx_stack[0].functions[funcref]
        funcref_list = list(ctx_stack[0].functions.keys())
        funcref.uniqueid = funcref_list[funcref_list.index(funcref)].uniqueid

    args = func_ctx.argdict(args, kwds)

    if not should_inline(func_ir, func_ctx, args):
        return ir.FunctionCall(operands=list(args.values()),
                               ret_dtype=func_ctx.ret_dtype,
                               name=funcref.name)
    else:
        return inline_expr(func_ir, func_ctx, args)


def create_in_intf(i, ctx):
    p = i.consumers[-1]
    intf_name = f'{p.gear.basename}_{p.basename}'
    # p.producer.source(HDLProducer())
    ir_intf = ir.Variable(intf_name, ir.IntfType[p.producer.dtype, ir.IntfType.iout])
    ctx.scope[intf_name] = ir_intf
    return ctx.ref(intf_name)


def create_out_intf(p, ctx):
    intf_name = f'{p.gear.basename}_{p.basename}'
    ir_intf = ir.Variable(intf_name, ir.IntfType[p.consumer.dtype, ir.IntfType.iin])
    ctx.scope[intf_name] = ir_intf
    p.consumer.connect(HDLConsumer())
    return ctx.ref(intf_name)


def call_gear(func, args, kwds, ctx: Context):
    local_in = []
    for i, a in enumerate(args):
        if typeof(a.dtype, ir.IntfType):
            if not isinstance(a, ir.Name):
                raise Exception(f'Expressions with interfaces not yet supported')

            intf = Intf(a.dtype.dtype)
            for p in ctx.gear.in_ports:
                if p.basename == a.name:
                    intf.source(p)
                    break
            else:
                raise Exception(f'Declaring free interfaces in async gears not yet supported')
        else:
            intf = Intf(a.dtype)
            intf.source(HDLProducer())

        local_in.append(intf)

    if not all(isinstance(node, ir.ResExpr) for node in kwds.values()):
        raise Exception("Not supproted")

    reg['gear/exec_context'] = 'compile'
    outputs = func(*local_in, **{k: v.val for k, v in kwds.items()})
    reg['gear/exec_context'] = 'hls'

    if isinstance(outputs, tuple):
        gear_inst = outputs[0].producer.gear
    else:
        gear_inst = outputs.producer.gear

    in_ports = []
    for a, i in zip(args, local_in):
        if typeof(a.dtype, ir.IntfType):
            in_ports.append(a)
            continue

        in_ports.append(create_in_intf(i, ctx))

    stmts = []
    for a, intf in zip(args, in_ports):
        if a == intf:
            continue

        stmts.append(ir.AssignValue(ir.Component(intf, 'data'), a))

    # TODO: Hack! This functionality needs to be rewriten to resemble the way
    # hierarchical modules work. This only works if ccat is automatically
    # placed in front of a gear.
    in_gear_inst = local_in[0].consumers[-1].gear
    if in_gear_inst is not gear_inst:
        ccat_out = create_out_intf(in_gear_inst.out_ports[0], ctx)
        ctx.submodules.append(Submodule(in_gear_inst, in_ports, [ccat_out]))

        in_ports = [ccat_out]

    out_ports = []
    for p in gear_inst.out_ports:
        out_ports.append(create_out_intf(p, ctx))

    ctx.submodules.append(Submodule(gear_inst, in_ports, out_ports))

    if len(out_ports) == 1:
        return ctx.ref(out_ports[0].name), stmts
    else:
        return ir.TupleExpr(tuple(ctx.ref(p.name) for p in out_ports)), stmts



from pygears import alternative, TypeMatchError, gear
from pygears.typing import Union
from pygears.lib import fmap as common_fmap
from pygears.lib.mux import mux
from pygears.lib.demux import demux_ctrl
from pygears.lib.ccat import ccat
from pygears.lib.shred import shred


def unionmap_check(dtype, f, mapping):
    if not issubclass(dtype, Union):
        return False

    try:
        num_f = len(f)
    except TypeError:
        raise TypeMatchError(
            f'Union fmap argument "f" needs to be a sequence, received {f}')

    if mapping is None:
        num_types = len(list(dtype.types))
    else:
        num_types = max(mapping.values()) + 1

    if num_types != num_f:
        raise TypeMatchError(
            'Number of union types different from the number of fmap functions'
        )

    return True


@alternative(common_fmap)
@gear(enablement=b'unionmap_check(din, f, mapping)')
def unionmap(din,
             *,
             f,
             fdemux=demux_ctrl,
             fmux=mux,
             balance=None,
             mapping=None,
             use_dflt=True):

    if mapping:
        fdemux = fdemux(mapping=mapping)
        fmux = fmux(mapping=mapping)

    demux_dout = din | fdemux
    ctrl = demux_dout[0]
    branches = demux_dout[1:]

    dout = []
    for i, fd in enumerate(f):
        if fd is None:
            if balance is None:
                dout.append(branches[i])
            else:
                dout.append(branches[i] | balance)
        else:
            dout.append(fd(branches[i]))

        if dout[-1] is None or isinstance(dout[-1], tuple):
            ret = 'none' if dout[-1] is None else f'{len(dout[-1])} outputs'
            raise TypeMatchError(
                f'Gear "{fd}" passed to the unionmap should have a single output, but returned {ret}'
            )

    # Situation where there is a default branch because of mapping
    if len(branches) == len(dout) + 1 and mapping is not None:
        if use_dflt:
            dout.append(branches[-1])
        else:
            branches[-1] | shred

    elif len(branches) > len(dout):
        raise Exception

    if balance is not None:
        ctrl = ctrl | balance

    if len(dout) == 1:
        return ccat(*dout, ctrl) | Union
    else:
        return fmux(ctrl, *dout)



from pygears import gear, alternative
from pygears.typing import Queue, Uint, Bool


@gear
async def group(din: Queue, size: Uint, *,
                init=1) -> Queue['din.data', 'din.lvl + 1']:

    cnt = size.dtype(init)
    last: Bool
    out_eot: Uint[din.dtype.lvl+1]

    async with size as c:
        assert c >= init, 'group: incorrect configuration'
        last = False
        while not last:
            async for (data, eot) in din:
                last = (cnt == c)
                out_eot = last @ eot
                yield (data, out_eot)
                if not last and all(eot):
                    cnt += 1


@alternative(group)
@gear
async def group_other(din, size: Uint, *, init=1) -> Queue['din']:
    cnt = size.dtype(init)
    last: Bool

    async with size as c:
        assert c >= init, 'group: incorrect configuration'
        last = False
        while not last:
            async with din as data:
                last = (cnt == c)
                yield (data, last)
                cnt += 1



import fnmatch
import os
import functools
from string import Template

from pygears import PluginBase, reg
from pygears.conf import Inject, inject, reg
from pygears.core.port import OutPort, InPort, HDLProducer
from .util import svgen_typedef
from pygears.hdl import hdlmod, rename_ambiguous

dti_spy_connect_t = Template("""
dti_spy #(${intf_name}_t) _${intf_name}(clk, rst);
assign _${intf_name}.data = ${conn_name}.data;
assign _${intf_name}.valid = ${conn_name}.valid;
assign _${intf_name}.ready = ${conn_name}.ready;""")


class SVIntfGen:
    def __init__(self, intf, lang=None):
        self.intf = intf
        self.lang = lang
        if self.lang is None:
            self.lang = reg['hdl/lang']

    @property
    @functools.lru_cache()
    def traced(self):
        def check(pattern):
            if isinstance(pattern, str):
                return fnmatch.fnmatch(self.intf.name, pattern)
            else:
                return pattern(self.intf)

        return any(check(p) for p in reg['debug/trace'])

    @property
    @functools.lru_cache(maxsize=None)
    def _basename(self):
        return self.intf.basename

    @property
    def parent(self):
        prod_port = self.intf.producer

        if isinstance(prod_port, InPort):
            return prod_port.gear

        return prod_port.gear.parent

    @property
    @functools.lru_cache(maxsize=None)
    def basename(self):
        basename = self._basename
        if self.is_port_intf:
            return basename

        cnt = 0
        for c in self.parent.child:
            if hdlmod(c)._basename == basename:
                cnt += 1

        for c in self.parent.local_intfs:
            if c is self.intf:
                break

            if hdlmod(c)._basename == basename:
                cnt += 1

        for p in self.parent.out_ports:
            if p.basename == basename:
                cnt += 1

        if cnt == 0:
            basename = f'{self._basename}_s'
        else:
            basename = f'{self._basename}{cnt}_s'

        return basename

    @property
    def is_port_intf(self):
        if isinstance(self.intf.producer, InPort):
            return True
        elif ((not self.is_broadcast) and self.intf.consumers
              and isinstance(self.intf.consumers[0], OutPort)):
            return True
        else:
            return False

    @property
    def sole_intf(self):
        if self.intf.producer:
            return len(self.intf.producer.gear.out_ports) == 1
        else:
            return True

    @property
    def is_broadcast(self):
        return len(self.intf.consumers) > 1

    @property
    def outname(self):
        if self.is_broadcast:
            return f'{self.basename}_bc'
        else:
            return self.basename

    @inject
    def get_inst(self, template_env, spy_connect_t=Inject('svgen/spy_connection_template')):

        if self.intf.producer is None:
            return

        inst = []
        if not self.is_port_intf:
            inst.append(self.get_intf_def(self.basename, 1, template_env))

        if self.is_broadcast:
            inst.extend([
                self.get_intf_def(self.outname, len(self.intf.consumers), template_env),
                self.get_bc_module(template_env)
            ])

            if self.traced and self.lang == 'sv':
                for i in range(len(self.intf.consumers)):
                    intf_name = f'{self.outname}_{i}'
                    conn_name = f'{self.outname}[{i}]'
                    if reg['debug/expand_trace_data']:
                        inst.extend(svgen_typedef(self.intf.dtype, intf_name).split('\n'))

                    inst.extend(
                        spy_connect_t.substitute(
                            intf_name=intf_name,
                            conn_name=conn_name,
                            width=self.intf.dtype.width,
                        ).split('\n'))

        for i, cons_port in enumerate(self.intf.consumers):
            if isinstance(cons_port, OutPort):
                if self.is_broadcast or isinstance(self.intf.producer, InPort):
                    din_name = self.outname
                    index = ''
                    if self.is_broadcast:
                        index = f'[{i}]'

                    inst.append(
                        template_env.snippets.intf_intf_connect(din_name,
                                                                self.intf.consumers[i].basename,
                                                                index))

        if self.traced and self.lang == 'sv':
            # TODO: Not a great abstraction. This basically differentiates
            # between Gearbox and Websim
            if reg['debug/expand_trace_data']:
                inst.extend(svgen_typedef(self.intf.dtype, self.basename).split('\n'))

                inst.extend(
                    spy_connect_t.substitute(intf_name=self.basename,
                                            conn_name=self.basename).split('\n'))
            else:
                inst.extend(
                    spy_connect_t.substitute(intf_name=f'{self.intf.basename}_spy',
                                             conn_name=self.basename,
                                             width=self.intf.dtype.width).split('\n'))

        return '\n'.join(inst)

    def get_intf_def(self, name, size, template_env):
        ctx = {
            'name': name,
            'width': self.intf.dtype.width,
            'size': size,
            'type': str(self.intf.dtype)
        }

        inst = []
        if self.lang == 'v' and self.traced:
            inst.append('/*verilator tracing_on*/')

        inst.append(template_env.snippets.intf_inst(**ctx))

        if self.lang == 'v' and self.traced:
            inst.append('/*verilator tracing_off*/')

        return '\n'.join(inst)

    def get_bc_module(self, template_env):
        inst_name = f'bc_{self.basename}'
        if inst_name.endswith('_if_s'):
            inst_name = inst_name[:-len('_if_s')]

        if self.lang == 'sv':
            bc_context = {
                'rst_name': 'rst',
                'module_name': rename_ambiguous('bc', self.lang),
                'inst_name': inst_name,
                'param_map': {
                    'SIZE': len(self.intf.consumers)
                },
                'port_map': {
                    'din': self.basename,
                    'dout': self.outname
                }
            }
        else:
            bc_context = {
                'rst_name': 'rst',
                'module_name': rename_ambiguous('bc', self.lang),
                'inst_name': inst_name,
                'param_map': {
                    'SIZE': len(self.intf.consumers),
                    'WIDTH': self.intf.dtype.width
                },
                'port_map': {
                    'din': (self.basename, None, None),
                    'dout': (self.outname, None, None)
                }
            }

        return template_env.snippets.module_inst(**bc_context)


class SVGenIntfPlugin(PluginBase):
    @classmethod
    def bind(cls):
        reg['svgen/spy_connection_template'] = dti_spy_connect_t



from pygears import alternative, gear
from pygears.lib import cart
from pygears.typing import Bool, Queue, Tuple, Uint


@gear
async def take(din: Queue[Tuple[{
        'data': 't_data',
        'size': Uint
}], 't_lvl'],
               *,
               init=1) -> Queue['t_data', 't_lvl']:
    """Takes the requested number of input data. The rest are consumed. The
    ``data`` field of the :class:`Tuple` type is passed to output, while the
    ``size`` field specifies the requested count. The module resets when the
    input transaction is finished.

    Args: init: Initialization value for the counter

    Returns: A :class:`Queue` type whose data consists of the input ``data``
        field of the :class:`Tuple`. """

    cnt = din.dtype.data['size'](init)
    last_take: Bool

    async for ((data, size), eot) in din:
        last_take = (cnt == size)

        if (cnt <= size):
            yield (data, eot | (last_take << (din.dtype.lvl - 1)))

        if all(eot[:(din.dtype.lvl - 1)]):
            cnt += 1


@alternative(take)
@gear
def take2(din: Queue, size: Uint):
    return cart(size, din, order=[1, 0]) | take



from pygears import gear, reg
from pygears.typing import Union
from pygears.lib.demux import demux_ctrl
from pygears.lib.mux import mux
from pygears.lib.shred import shred


def fill_type(din_t, union_t, sel):
    dtypes = list(union_t.types).copy()
    dtypes[sel] = din_t

    return Union[tuple(dtypes)]


@gear
def fill(din,
         union_din: Union,
         *,
         fdemux=demux_ctrl,
         fmux=mux,
         sel) -> b'fill_type(din, union_din, sel)':
    fields = union_din | fdemux
    fields_list = list(fields)

    fields_list[sel+1] | shred

    fields_list[sel+1] = din
    return tuple(fields_list) | fmux



from .fmap import fmap
from pygears.lib.fmaps import unionmap, queuemap, tuplemap, arraymap

from .expand import expand
from .factor import factor
from .flatten import flatten
from .project import project
from .czip import czip, zip_sync, unzip, zip_sync_with, zip_wrap_with
from .ccat import ccat, ccat_sync_with
from .cart import cart, cart_sync, cart_sync_with, cart_wrap_with
from .union import (union_sync, union_collapse, union_sync_with, case, when,
                    mux_by, ucase, select, field_sel, maybe_when)
from .queue import queue_wrap_from, sot_queue
from .cast import cast, trunc
from .quenvelope import quenvelope
from .sieve import sieve
from .permute import permuted_apply
from .const import const, fix, once, ping, void
from .rom import rom
from .operators import (add, div, eq, ge, gt, invert, le, lt, mod, mul, ne,
                        neg, sub, shl, shr, xor, code, or_, and_)
from .saturate import saturate
from .state import state
from .mux import mux, mux_zip, field_mux
from .demux import demux, demux_zip, demux_by, demux_ctrl
from .shred import shred
from .decouple import decouple, buff
from .fifo import fifo
from .fill import fill
from .dreg import dreg, regout
from .serialize import serialize
from .parallelize import parallelize
from .data_dly import data_dly
from .filt import filt
from .align import align
from .replicate import replicate, replicate_while, replicate_until
from .chop import chop
from .clip import clip
from .din_cat import din_cat
from .form_sub_cfg import form_sub_cfg
from .iceil import iceil
from .priority_mux import priority_mux
from .qcnt import qcnt
from .release_after_eot import release_after_eot
from .reverse import reverse
from .rng import qrange
from .sdp import sdp
from .tdp import tdp
from .take import take
from .group import group
from .interlace import qinterlace, interlace
from .deal import qdeal, deal
from .repack import repack
from .priority_mux_valve import priority_mux_valve
from .max_gears import max2
from .unary import unary
from .alternate_queues import alternate_queues
from .delay import delay, delay_rng, delay_gen
from .pipeline import pipeline
from .verif import directed, verif, drv, check, mon, scoreboard, collect
from .drvrnd import drvrnd
from .reduce2 import reduce2
from .reduce import reduce, accum, qmax
from .funclut import funclut
from .rounding import truncate, round_half_up, round_to_zero, round_to_even, qround
from .scope import scope
from .asyncreg import trigreg, sample, regmap, avail
from .dispatch import dispatch

import pygears.lib.hls

__all__ = [
    'expand', 'czip', 'ccat', 'cart', 'cast', 'trunc', 'code', 'quenvelope',
    'zip_sync', 'sieve', 'flatten', 'fmap', 'unionmap', 'queuemap', 'tuplemap',
    'arraymap', 'permuted_apply', 'const', 'once', 'mux', 'field_mux', 'demux',
    'shred', 'cart_sync', 'decouple', 'dreg', 'regout', 'unzip', 'serialize',
    'project', 'fifo', 'factor', 'fill', 'mux_zip', 'demux_zip', 'demux_ctrl',
    'data_dly', 'union_sync', 'union_collapse', 'ucase', 'filt', 'buff',
    'queue_wrap_from', 'demux_by', 'mux_by', 'align', 'cart_sync_with', 'when',
    'union_sync_with', 'select', 'field_sel', 'case', 'fix', 'xor', 'rng',
    'qrange', 'priority_mux', 'qcnt', 'sdp', 'tdp', 'chop', 'qinterlace',
    'interlace', 'replicate', 'replicate_while', 'replicate_until', 'qdeal',
    'deal', 'clip', 'din_cat', 'take', 'release_after_eot', 'reverse',
    'form_sub_cfg', 'reduce2', 'group', 'repack', 'priority_mux_valve', 'max2',
    'unary', 'alternate_queues', 'delay', 'delay_rng', 'delay_gen', 'directed',
    'verif', 'collect', 'rom', 'drv', 'drvrnd', 'check', 'mon', 'reduce',
    'accum', 'pipe', 'funclut', 'truncate', 'round_half_up', 'round_to_zero',
    'round_to_even', 'add', 'div', 'eq', 'ge', 'gt', 'invert', 'iceil', 'le',
    'lt', 'mod', 'mul', 'ne', 'neg', 'sub', 'shl', 'shr', 'xor', 'scoreboard',
    'scope', 'sample', 'trigreg', 'regmap', 'saturate', 'qround',
    'parallelize', 'ping', 'state', 'avail'
]



""":class:`Union` is a generic type somewhat similar to the unions in other
HDLs and C language for an example, with an important differences. Unions in
these languages only act as different views to the underlying data, allowing
user to read the same data in different formats, called subtypes here. However,
the data of the PyGears :class:`Union` type can represent only one of its
subtypes, and it carries the information which of the subtypes it represents.

A concrete :class:`Union` type is created by listing the subtypes that it can
represent. For an example a :class:`Union` type that represents either a
an integer or a fixed-point number can be specified like this::

    FixpInt = Union[Uint[16], Tuple[Uint[8], Uint[8]]]

This is same as if the tuple of types were supplied::

    fields = (Uint[16], Tuple[Uint[8], Uint[8]])
    FixpInt = Union[fields]

If we want to create a type that is generic in terms of the integer width, we
can provide a template parameter name instead of the concrete widths::

    FixpInt = Union[Uint['w_int'], Tuple[Uint['w_fixp_q'], Uint['w_fixp_p']]]


``FixpInt`` is now a generic template type, represented by a :class:`Union`
with two subtypes, which is not yet fully specified. The ``FixpInt`` template
has a three parameters (``w_int``, ``w_fixp_q`` and ``w_fixp_p``), which need
to be specified to get a concrete type. So if we want to obtain a ``FixpInt``
with 16 bit integer type and Q8.8 fixed-point type, we can write::

    FixpIntQ8_8 = FixpInt[16, 8, 8]

We can also be explicit which template parameter is assigned a concrete type::

    FixpIntQ8_8 = FixpInt[{
        'w_int':    16,
        'w_fixp_q': 8,
        'w_fixp_p': 8
    }]

Once a concrete type has been formed it can be instantiated, which is useful
for the verification. Type instance is obtained by specifying two arguments: a
value and and the ID of the active subtype. The ID of the subtypes is its index
in the subtype list provided when defining the :class:`Union`::

    uint_val = FixpIntQ8_8(0xbeef, 0)
    fixp_val = FixpIntQ8_8((0xbe, 0xef), 1)

We can now check the contents of the created data, via their ``data`` and
``ctrl`` fields:

>>> uint_val.data
Uint[16](48879)
>>> uint_val.ctrl
Uint[1](0)

>>> fixp_val.data
(Uint[8](190), Uint[8](239))
>>> fixp_val.ctrl
Uint[1](1)

"""

from .base import EnumerableGenericMeta, type_str, is_type
from .base import class_and_instance_method, TemplatedTypeUnspecified
from .math import bitw
from .unit import Unit
from .uint import Uint


class UnionType(EnumerableGenericMeta):
    """Implements the :class:`Union` generic type.

    All operations on the :class:`Union` type are implemented here in the
    :class:`UnionType` class. Operations on the :class:`Union` type instances
    are defined in the :class:`Union` class.
    """

    # def __new__(cls, name, bases, namespace, args=[]):
    #     cls = super().__new__(cls, name, bases, namespace, args)

    #     args = cls.args
    #     if not args:
    #         return cls
    #     elif len(args) == 0:
    #         return Unit
    #     elif len(args) == 1:
    #         return args[0]
    #     else:
    #         return cls

    def __getitem__(self, parameters):
        if not self.specified:
            return super().__getitem__(parameters)

        index = self.index_norm(parameters)

        if isinstance(index[0], slice):
            if (index[0].start == 0) and (index[0].stop == 1):
                return Uint[max(a.width for a in self.args)]
            elif (index[0].start == 1) and (index[0].stop == 2):
                return Uint[bitw(len(self.args) - 1)]
            elif (index[0].start == 0) and (index[0].stop == 2):
                return Uint[max(a.width for a in self.args)], Uint[bitw(len(self.args) - 1)]
            else:
                raise IndexError
        else:
            if (index[0] == 0):
                return Uint[max(a.width for a in self.args)]
            elif (index[0] == 1):
                return Uint[bitw(len(self.args) - 1)]
            else:
                raise IndexError

    @property
    def fields(self):
        return ['data', 'ctrl']

    def keys(self):
        return [0, 1]

    @property
    def data(self):
        '''Returns the type of the :class:`Union` ``data`` field. This field is of type
        :class:`Uint`, large enough to contain the largest subtype of the
        :class:`Union`.

        '''
        return self[0]

    @property
    def ctrl(self):
        '''Returns the type of the :class:`Union` ``ctrl`` field, which contains the
        index of the subtype represented by a :class:`Union` instance. This
        field is of type :class:`Uint`, large enough to contain the all
        :class:`Union` subtype indices.

        '''
        return self[-1]

    @property
    def types(self):
        '''Returns a list of subtypes.'''
        return self.args

    @property
    def width(self):
        return sum(f.width for f in self)

    def __str__(self):
        return '%s' % ' | '.join([type_str(a) for a in self.args])


class Union(tuple, metaclass=UnionType):
    def __new__(cls, val=None, ctrl=None):
        if type(val) == cls:
            return val

        if not cls.specified:
            raise TemplatedTypeUnspecified

        if val is None:
            val = cls.types[0](val)
            ctrl = 0

        if ctrl is None:
            val, ctrl = val

        subtype = cls.types[ctrl]
        data_type = cls[0]

        # TODO: This allows for coded value to be supplied to union. This
        # should probably be refactored into different API
        if type(val) == data_type:
            data = val
        else:
            try:
                subval = subtype(val)
                data = data_type(subval.code())
            except TypeError as e:
                raise TypeError(f'{str(e)}\n - when instantiating subtype "{repr(subtype)}"'
                                f' with "{repr(val)}"')

        return super(Union, cls).__new__(cls, (data, cls[1](ctrl)))

    def __int__(self):
        """Returns a packed integer representation of the :class:`Union` instance.
        """
        ret = 0

        for d, t in zip(reversed(self), reversed(type(self))):
            ret <<= t.width
            ret |= int(d)

        return ret

    def __eq__(self, other):
        if not is_type(type(other)):
            return super().__eq__(other)

        return type(self) == type(other) and super().__eq__(other)

    def __hash__(self):
        return super().__hash__()

    def __ne__(self, other):
        if not is_type(type(other)):
            return super().__ne__(other)

        return not self.__eq__(other)

    def code(self):
        """Returns a packed integer representation of the :class:`Union` instance.
        """
        ret = 0

        for d, t in zip(reversed(self), reversed(type(self))):
            ret <<= t.width
            ret |= d.code()

        return ret

    @class_and_instance_method
    @property
    def data(self):
        """Returns the data carried by the :class:`Union` instance, converted to the
        represented subtype."""
        # return type(self).types[self[1]].decode(self[0])
        return self[0]

    @class_and_instance_method
    @property
    def ctrl(self):
        """Returns the index of the subtype represented by a :class:`Union`
        instance."""
        return self[-1]

    @classmethod
    def decode(cls, val):
        """Returns a :class:`Union` instance from its packed integer representation.
        """

        ret = []
        val = int(val)
        for t in cls:
            t_width = t.width
            t_mask = (1 << t_width) - 1
            ret.append(val & t_mask)
            val >>= t_width

        data, ctrl = ret
        subtype = cls.types[ctrl]

        return cls((subtype.decode(data), cls[1].decode(ctrl)))


class classproperty(object):
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


class Maybe(Union):
    def __new__(cls, val=None):
        if type(val) == cls:
            return val

        if not cls.specified:
            raise TemplatedTypeUnspecified

        if val is None:
            val = cls.types[0]()
            ctrl = 0
        else:
            val, ctrl = val

        if ctrl == 0:
            return super(Union, cls).__new__(cls, (cls[0](cls.dtype().code()), cls[1](0)))
        elif type(val) == cls[0]:
            data = val
        else:
            subtype = cls.types[ctrl]
            data_type = cls[0]

            try:
                subval = subtype(val)
                data = data_type(subval.code())
            except TypeError as e:
                raise TypeError(f'{str(e)}\n - when instantiating subtype "{repr(subtype)}"'
                                f' with "{repr(val)}"')

        return super(Union, cls).__new__(cls, (data, cls[1](1)))

    def get(self):
        return self.dtype.decode(self.data)

    def __bool__(self):
        return bool(self.ctrl)

    @classmethod
    def some(cls, val):
        return cls((val, True))

    @classproperty
    def dtype(cls):
        return cls.types[1]


# TODO: typeof does not work correctly
Maybe = Maybe[Unit, 'data']

Nothing = Maybe[Uint[1]]()

def some(v):
    return Maybe[type(v)]((v, True))



import inspect
import os

from pygears import PluginBase, reg
from pygears.core.hier_node import HierVisitorBase
from .intf import VIntfGen
from pygears.definitions import LIB_VLIB_DIR, USER_VLIB_DIR


class VGenInstVisitor(HierVisitorBase):
    def __init__(self):
        self.namespace = reg['vgen/module_namespace']
        self.vgen_map = reg['vgen/map']

    def RTLGear(self, node):
        if 'hdl' not in node.meta_kwds:
            node.meta_kwds['hdl'] = {}

        if 'vgen_cls' not in node.meta_kwds['hdl']:
            vgen_cls = self.namespace.get(node.gear.definition, None)

            if vgen_cls is None:
                for base_class in inspect.getmro(node.gear.__class__):
                    if base_class.__name__ in self.namespace:
                        vgen_cls = self.namespace[base_class.__name__]
                        break

            node.meta_kwds['hdl']['vgen_cls'] = vgen_cls

    def RTLNode(self, node):
        vgen_cls = node.meta_kwds['hdl']['vgen_cls']

        if vgen_cls:
            svgen_inst = vgen_cls(node)
        else:
            svgen_inst = None

        self.vgen_map[node] = svgen_inst

    def RTLIntf(self, intf):
        self.vgen_map[intf] = VIntfGen(intf)


def vgen_inst(top, conf):
    v = VGenInstVisitor()
    v.visit(top)

    return top


def vgen_include_get(cfg):
    return reg['hdl/include'] + [USER_VLIB_DIR, LIB_VLIB_DIR]


class VGenInstPlugin(PluginBase):
    @classmethod
    def bind(cls):
        pass
        # reg['vgen/map'] = {}
        # reg['vgen/module_namespace'] = {}
        # reg.confdef('vgen/include', getter=vgen_include_get)

    @classmethod
    def reset(cls):
        reg['vgen/map'] = {}



from pygears import reg
from pygears.hdl.sv import SVGenPlugin
from pygears.hdl.sv.svmod import SVModuleInst
from pygears.lib.align import align


class SVGenAlign(SVModuleInst):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):
        params = {}

        from pygears.typing import Queue
        for l, p in zip(self.node.params['lvl'], self.node.in_ports):
            dtype = p.dtype
            if issubclass(dtype, Queue):
                dtype = dtype[0]
            params[f'W_{p.basename.upper()}_LVL'] = l
            params[f'W_{p.basename.upper()}_DATA'] = dtype.width

        context = {
            'module_name': self.module_name,
            'params': params,
            'intfs': list(self.port_configs)
        }
        return template_env.render_local(__file__, "align.j2", context)


class SVGenAlignPlugin(SVGenPlugin):
    @classmethod
    def bind(cls):
        reg['svgen/module_namespace'][align] = SVGenAlign



from pygears import gear


@gear
async def reverse(din: 'TDin') -> b'TDin':
    """Bit reversal
    for example: 65 → 01000001 → 10000010 → 130 (for TDin = Uint[8])
    """
    async with din as data:
        b = '{:0{width}b}'.format(data, width=len(din.dtype))
        yield int(b[::-1], 2)



from pygears import gear, alternative
from pygears import datagear
from pygears.typing import Queue, Union, typeof, Tuple, Uint, Any


def mux_type(dtypes, mapping):
    full_dtypes = tuple(dtypes[mapping[i]] for i in sorted(mapping))
    return Union[full_dtypes]


def full_mapping(dtypes, mapping):
    dout_num = max(mapping.values()) + 1
    fm = mapping.copy()

    for i in range(len(dtypes)):
        if i not in mapping:
            fm[i] = dout_num

    return fm


def dflt_map(dtypes):
    return {i: i for i in range(len(dtypes))}


@gear
async def mux(
        ctrl: Uint,
        *din,
        mapping=b'dflt_map(din)',
        _full_mapping=b'full_mapping(din, mapping)') -> b'mux_type(din, _full_mapping)':

    async with ctrl as c:
        c_map = _full_mapping[int(c)]

        assert c_map < len(din), 'mux: incorrect selection value'
        async with din[c_map] as d:
            yield (d, c)


@gear
async def mux_zip(
        ctrl, *din, mapping=b'dflt_map(din)',
        _full_mapping=b'full_mapping(din, mapping)') -> b'mux_type(din, _full_mapping)':
    async with ctrl as c:
        c_map = _full_mapping[int(c)]

        assert c_map < len(din), 'mux: incorrect selection value'
        async for d in din[c_map]:
            yield (d, c)


@datagear
def field_mux(
        din: Tuple[{
            'ctrl': Uint,
            'data': Any
        }],
        *,
        mapping=b'dflt_map(din["data"])',
        _full_mapping=b'full_mapping(din["data"], mapping)'
) -> b'mux_type(din["data"], _full_mapping)':
    c_map = _full_mapping[int(din['ctrl'])]
    return (din['data'][c_map], c_map)



from pygears import reg
from pygears.hdl.sv.svmod import SVModuleInst
from pygears.hdl.sv import SVGenPlugin
from pygears.lib.demux import demux_zip
from .syncguard import SVGenSyncGuard


class SVGenDemuxZip(SVModuleInst):
    def __init__(self, node):
        super().__init__(node)
        self.syncguard = SVGenSyncGuard(f'{self.module_name}_syncguard', 2)

    @property
    def is_generated(self):
        return True

    @property
    def file_name(self):
        return super().file_name, self.syncguard.file_name

    def get_module(self, template_env):
        context = {
            'module_name': self.module_name,
            'intfs': list(self.port_configs)
        }
        return template_env.render_local(
            __file__, "demux_zip.j2", context), \
            self.syncguard.get_module(template_env)


class SVGenDemuxZipPlugin(SVGenPlugin):
    @classmethod
    def bind(cls):
        reg['svgen/module_namespace'][demux_zip] = SVGenDemuxZip



import os

import jinja2

from pygears.conf import reg
from pygears.hls import HDLVisitor
from pygears.typing import Queue, typeof, code

from .util import vgen_intf, vgen_signal
from .v_expression import cast, vexpr

REG_TEMPLATE = """
always @(posedge clk) begin
    if(rst | _rst_cond) begin
        {0}_reg <= {1};
    end else if ({0}_en) begin
        {0}_reg <= {0}_next;
    end
end
"""


class HDLWriter:
    def __init__(self):
        self.indent = 0
        self.lines = []

    def line(self, line=''):
        if not line:
            self.lines.append('')
        else:
            self.lines.append(f'{" "*self.indent}{line}')

    def block(self, block):
        for line in block.split('\n'):
            self.line(line)


class VCompiler(HDLVisitor):
    def __init__(self, visit_var, writer, hdl_locals, **kwds):
        self.writer = writer
        self.visit_var = visit_var
        self.hdl_locals = hdl_locals
        self.extras = {}
        self.kwds = kwds

        self.separated = kwds.get('separated_visit', False)
        self.condtitions = kwds['conditions']

    def resolve_stmt(self, node):
        target = vexpr(node.target, extras=self.extras)

        if self.separated:
            if target != self.visit_var:
                return

        val = node.val
        if isinstance(val, str) and val in self.condtitions:
            val = self.condtitions[val]

        rhs = vexpr(val, extras=self.extras)

        var = None
        if target in self.hdl_locals:
            var = self.hdl_locals[target]

        if node.dtype or var is None:
            return f'{target} = {rhs};'

        if var.dtype.width == node.dtype.width:
            return f'{target} = {cast(var.dtype, node.val.dtype, rhs)};'

        assert False, 'node.dtype diff from hdl local width'
        return None

    def enter_block(self, block):
        if getattr(block, 'in_cond', False):
            in_cond = block.in_cond
            if isinstance(in_cond, str) and in_cond in self.condtitions:
                in_cond = self.condtitions[in_cond]
            self.writer.line(
                f'if ({vexpr(in_cond, extras=self.extras)}) begin')

        if getattr(block, 'in_cond', True):
            self.writer.indent += 4

    def exit_block(self, block=None):
        if getattr(block, 'in_cond', True):
            self.writer.indent -= 4
            self.writer.line(f'end')

    def visit_AssertValue(self, node):
        if 'formal' in self.kwds and self.kwds['formal']:
            self.writer.line(
                f'assume ({vexpr(node.val.test, extras=self.extras)});')
        else:
            self.writer.line(
                f'if (!({vexpr(node.val.test, extras=self.extras)})) begin')
            self.writer.indent += 4
            self.writer.line(f'$display("{node.val.msg}");')
            # self.writer.line(f'$finish;')
            self.writer.indent -= 4
            self.writer.line(f'end')

    def visit_AssignValue(self, node):
        val = self.resolve_stmt(node)
        if val is not None:
            self.writer.line(val)

    def visit_ModBlock(self, node):
        if not node.stmts and not node.dflts:
            return
        self.writer.line(f'// Comb block for: {self.visit_var}')
        self.writer.line(f'always @* begin')

        self.visit_HDLBlock(node)

        self.writer.line('')

    def visit_FuncReturn(self, node):
        self.writer.line(
            f"{vexpr(node.func.name)} = {vexpr(node.expr, extras=self.extras)};"
        )

    def visit_FuncBlock(self, node):
        size = ''
        if node.ret_dtype.width > 0:
            size = f'[{node.ret_dtype.width-1}:0]'

        if getattr(node.ret_dtype, 'signed', False):
            size = f'signed {size}'

        self.writer.line(f'function {size} {node.name};')

        sigdef = {}

        self.writer.indent += 4

        for name, arg in node.args.items():
            # size = ''
            # if len(arg.dtype) > 0:
            #     size = f'[{len(arg.dtype)-1}:0]'

            #     if getattr(arg.dtype, 'signed', False):
            #         size = f'signed {size}'

            arg_name = vexpr(arg)

            self.writer.line(
                vgen_signal(arg.dtype, 'input', arg_name, 'input', False))

            sigdef[arg_name] = vgen_signal(arg.dtype, 'reg', arg_name, 'input',
                                           True).split('\n')[1:]

            for l in sigdef[arg_name]:
                if l.startswith('reg'):
                    self.writer.line(l)

        for name, expr in node.hdl_data.variables.items():
            if name not in node.args:
                self.writer.block(
                    vgen_signal(expr.dtype, 'reg', f'{name}_v', 'input'))

        self.writer.indent -= 4

        if not node.stmts and not node.dflts:
            return

        self.writer.line(f'begin')

        self.writer.indent += 4
        for name, sdef in sigdef.items():
            for l in sdef:
                if l.startswith('assign'):
                    self.writer.line(l[6:])

        self.writer.indent -= 4

        self.visit_HDLBlock(node)

        self.writer.line(f'endfunction')
        self.writer.line('')

    def visit_CombSeparateStmts(self, node):
        if node.stmts:
            self.writer.line(f'// Comb statements for: {self.visit_var}')
            for stmt in node.stmts:
                val = self.resolve_stmt(stmt)
                if val is not None:
                    self.writer.line(f'assign {val}')
            self.writer.line('')

    def visit_HDLBlock(self, node):
        self.enter_block(node)

        for stmt in node.dflt_stmts:
            val = self.resolve_stmt(stmt)
            if val is not None:
                self.writer.line(val)

        for stmt in node.stmts:
            self.visit(stmt)

        self.exit_block(node)


DATA_FUNC_GEAR = """
{%- import 'snippet.j2' as snippet -%}

{% call snippet.module_with_intf_structs(module_name, intfs, intfs, comment) %}

{{vlines|indent(4,True)}}

{%- endcall %}
"""


def write_module(hdl_data, v_stmts, writer, **kwds):
    if 'config' not in kwds:
        kwds['config'] = {}

    extras = {}

    separate_conditions(v_stmts, kwds, vexpr)

    for name, expr in hdl_data.hdl_functions.items():
        compiler = VCompiler(name, writer, hdl_data.hdl_locals, **kwds)
        compiler.visit(expr)
        extras.update(compiler.extras)

    for name, expr in hdl_data.regs.items():
        writer.line(
            vgen_signal(expr.dtype, 'reg', f'{name}_reg', 'input', False))
        writer.line(
            vgen_signal(expr.dtype, 'reg', f'{name}_next', 'input', False))
        writer.line(f'reg {name}_en;')
        writer.line()

    for name, val in hdl_data.in_intfs.items():
        writer.line(vgen_intf(val.dtype, name, 'input', False))
        writer.line(vgen_signal(val.dtype, 'reg', f'{name}_s', 'input', False))
        tmp = vgen_signal(val.dtype, 'wire', f'{name}_s', 'input')
        writer.line(tmp.split(';', 1)[1])
        writer.line(f"assign {name} = {name}_s;")
    writer.line()

    for name, expr in hdl_data.variables.items():
        writer.block(
            vgen_signal(expr.dtype, 'reg', f'{name}_v', 'input', False))
        writer.line()

    if 'conditions' in v_stmts:
        for cond in v_stmts['conditions'].stmts:
            writer.line(f'wire {cond.target};')
        writer.line()

    if hdl_data.regs:
        writer.line(f'initial begin')
        for name, expr in hdl_data.regs.items():
            writer.line(f"    {name}_reg = {int(code(vexpr(expr.val)))};")

        writer.line(f'end')

    for name, expr in hdl_data.regs.items():
        writer.block(REG_TEMPLATE.format(name, int(code(vexpr(expr.val)))))

    for name, val in v_stmts.items():
        if name != 'variables':
            compiler = VCompiler(name, writer, hdl_data.hdl_locals, **kwds)
            compiler.visit(val)
            extras.update(compiler.extras)
        else:
            kwds['separated_visit'] = True
            for var_name in hdl_data.variables:
                compiler = VCompiler(f'{var_name}_v', writer,
                                     hdl_data.hdl_locals, **kwds)
                compiler.visit(val)
                extras.update(compiler.extras)
            kwds['separated_visit'] = False

    for func_impl in extras.values():
        writer.block(func_impl)


def write_assertions(gear, writer, cfg):
    asserts = cfg['asserts'] if 'asserts' in cfg else {}
    assumes = cfg['assumes'] if 'assumes' in cfg else []

    def append_to_context(context, port):
        if typeof(port.dtype, Queue):
            lvl = port.dtype.lvl
        else:
            lvl = 0

        tmp_asrt = []
        for key, val in asserts.items():
            if port.basename == key:
                tmp_asrt.append(val)

        context.append((port.basename, lvl, tmp_asrt))

    in_context = []
    for port in gear.in_ports:
        append_to_context(in_context, port)

    out_context = []
    for port in gear.out_ports:
        append_to_context(out_context, port)

    base_addr = os.path.dirname(__file__)
    jenv = jinja2.Environment(loader=jinja2.FileSystemLoader(base_addr),
                              trim_blocks=True,
                              lstrip_blocks=True)

    context = {
        'in_context': in_context,
        'out_context': out_context,
        'assumes': assumes
    }
    res = jenv.get_template('formal.j2').render(context)

    writer.block(res)


def compile_gear_body(gear):
    formal = False
    conf = reg['vgen']
    if 'formal' in conf:
        formal = conf['formal']

    hdl_data, res = parse_gear_body(gear)
    writer = HDLWriter()
    write_module(hdl_data,
                 res,
                 writer,
                 formal=formal,
                 config=gear.meta_kwds.get('hdl', {}))

    if formal:
        write_assertions(gear, writer, formal)

    return '\n'.join(writer.lines)


def compile_gear(gear, template_env, context):
    context['vlines'] = compile_gear_body(gear)
    return template_env.render_string(DATA_FUNC_GEAR, context)



import ast
from . import Context, HLSSyntaxError, ir, node_visitor, visit_ast, ir_utils
from pygears import Intf
from .stmt import infer_targets


def is_intf_list(node):
    if isinstance(node, ir.ConcatExpr):
        return all(ir_utils.is_intf_id(v) for v in node.operands)

    if not isinstance(node, ir.ResExpr):
        return False

    if not isinstance(node.val, list):
        return False


def parse_generator_expression(node, ctx):
    targets = visit_ast(node.target, ctx)

    out_intf_ref = visit_ast(node.iter, ctx)

    if is_intf_list(out_intf_ref):
        return out_intf_ref, targets, None

    gen_name = ctx.find_unique_name('_gen')
    ctx.scope[gen_name] = ir.Generator(gen_name, out_intf_ref)

    infer_targets(ctx, targets, out_intf_ref)

    return out_intf_ref, targets, gen_name


@node_visitor(ast.GeneratorExp)
def _(node: ast.GeneratorExp, ctx: Context):
    # out_intf_ref, targets, gen_name = parse_generator_expression(node.generators[0], ctx)

    # breakpoint()
    raise HLSSyntaxError(f"Unsupported language construct", node.lineno)



import inspect
import typing
from pygears.core.infer_ftypes import infer_ftypes
from pygears.core.gear import OutSig, InSig
from pygears.typing import Any, typeof
from functools import singledispatch
from dataclasses import dataclass
from .. import ir

from pygears import reg, Intf

from .utils import add_to_list, get_function_source, get_function_ast

from pygears.core.util import get_function_context_dict
from pygears.conf.trace import gear_definition_location
from pygears.conf.trace import make_traceback, TraceException
import sys


class HLSSyntaxError(TraceException):
    pass


def form_hls_syntax_error(ctx, e, lineno=1):
    msg = str(e)
    for c in reversed(reg['hls/ctx']):
        if isinstance(c, GearContext):
            func, fn, ln, _ = gear_definition_location(c.gear.func)
            msg += (f'\n    - when compiling gear "{c.gear.name}" with'
                    f' parameters {c.gear.params}')
        else:
            func, fn, ln, _ = gear_definition_location(c.funcref.func)
            msg += (f'\n    - when compiling function "{c.funcref.func}" with'
                    f' signature {c.signature}')

    err = HLSSyntaxError(msg, ln + lineno - 1, filename=fn)

    traceback = make_traceback((HLSSyntaxError, err, sys.exc_info()[2]))
    _, exc_value, tb = traceback.standard_exc_info

    return exc_value, tb


@dataclass
class Submodule:
    gear: typing.Any
    in_ports: typing.List
    out_ports: typing.List


def hashable(v):
    """Determine whether `v` can be hashed."""
    try:
        hash(v)
    except TypeError:
        return False
    return True


class Function:
    def __init__(self, func, args, kwds, uniqueid=None):
        self.source = get_function_source(func)
        self.func = func
        self.ast = get_function_ast(func)
        self.basename = ''.join(e for e in func.__name__ if e.isalnum() or e == '_')
        self.uniqueid = uniqueid

        # If there were some constant arguments, their values were embeded into
        # funtion code, so we need to take values as part of the hash.
        hargs = tuple(arg.val if isinstance(arg, ir.ResExpr) else arg.dtype for arg in args)

        hkwds = tuple((k, v) for k, v in kwds.items() if hashable(v))

        if self.source:
            # TODO: Include keywords here
            self._hash = hash(self.source) ^ hash(hargs) ^ hash(hkwds)
        else:
            self._hash = hash(self.__qualname__) ^ hash(hargs) ^ hash(hkwds)

    @property
    def name(self):
        if self.uniqueid is None:
            return self.basename

        return f'{self.basename}_{self.uniqueid}'

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __hash__(self):
        return self._hash


class Context:
    def __init__(self):
        self.scope: typing.Dict = {}
        self.methods = {}
        self.closures = []
        self.closure_alias_maps = {}
        self.expr_closure = None
        self.local_namespace: typing.Dict = {'__builtins__': __builtins__}
        self.ir_block_closure: typing.List = []
        self.submodules: typing.List[Submodule] = []
        self.reaching: typing.Dict = {}
        self.ast_stmt_map: typing.Dict = {}
        self.reaching = {}
        self.registers = {}

    def ref(self, name, ctx='load'):
        if name in self.scope:
            return ir.Name(name, self.scope[name], ctx=ctx)

        if name in self.local_namespace:
            return ir.ResExpr(self.local_namespace[name])

        raise KeyError

    def find_unique_name(self, name):
        res_name = name
        i = 0
        while res_name in self.scope:
            i += 1
            res_name = f'{name}_{i}'

        return res_name

    def new_closure(self, node):
        if self.closures:
            scope = self.alias_map.copy()
        else:
            scope = self.local_namespace.copy()

        self.closure_alias_maps[node] = scope
        self.closures.append(node)

    @property
    def alias_map(self):
        if self.closures:
            return self.closure_alias_maps[self.ir_parent_block]
        else:
            return {}

    @property
    def ir_parent_block(self):
        return self.closures[-1]

    @property
    def variables(self):
        return {
            name: obj
            for name, obj in self.scope.items() if isinstance(obj, ir.Variable) and obj.reg is False
        }

    @property
    def regs(self):
        return {}


class IntfProxy(Intf):
    def __init__(self, port):
        self.port = port

    @property
    def dtype(self):
        return self.port.dtype

    def __str__(self):
        return self.port.basename

    def __repr__(self):
        return repr(self.port)


class GearContext(Context):
    def __init__(self, gear):
        super().__init__()
        self.gear = gear
        self.functions: typing.Mapping[Function, FuncContext] = {}
        self.local_namespace = get_function_context_dict(self.gear.func).copy()

        paramspec = inspect.getfullargspec(self.gear.func)

        # TODO: If local variable shadows output interface name, they will clash!!! (Example: dout)
        vararg = []
        for p in self.gear.in_ports:
            self.scope[p.basename] = ir.Variable(p.basename, ir.IntfType[p.dtype, ir.IntfType.iin])

            if paramspec.varargs and p.basename.startswith(paramspec.varargs):
                vararg.append(self.ref(p.basename))

        if paramspec.varargs:
            self.local_namespace[paramspec.varargs] = ir.ConcatExpr(vararg)

        for p in self.gear.out_ports:
            self.scope[f'{p.basename}'] = ir.Variable(
                f'{p.basename}', ir.IntfType[p.consumer.dtype, ir.IntfType.iout])

        for k, v in self.gear.explicit_params.items():
            self.local_namespace[k] = ir.ResExpr(v)

    @property
    def intfs(self):
        return {
            name: obj
            for name, obj in self.scope.items()
            if isinstance(obj, ir.Variable) and typeof(obj.dtype, ir.IntfType)
        }

    @property
    def signals(self):
        return {
            name: obj
            for name, obj in self.scope.items()
            if isinstance(obj, ir.Variable) and isinstance(obj.val, (OutSig, InSig))
        }

    @property
    def regs(self):
        return {
            name: obj
            for name, obj in self.scope.items() if isinstance(obj, ir.Variable) and obj.reg
        }

    @property
    def in_ports(self):
        return [
            obj for obj in self.scope.values()
            if (isinstance(obj, ir.Interface) and isinstance(obj.intf, Intf) and obj.intf.producer
                and obj.intf.producer.gear is self.gear)
        ]

    @property
    def out_ports(self):
        return [self.ref(f'{p.basename}') for p in self.gear.out_ports]


class FuncContext(Context):
    def argdict(self, args, kwds):
        paramspec = inspect.getfullargspec(self.funcref.func)
        args = dict(zip(paramspec.args, args))
        args.update(kwds)

        for name in self.const_args:
            if name not in args:
                continue

            del args[name]

        return args

    def __init__(self, funcref: Function, args, kwds):
        super().__init__()
        self.funcref = funcref

        func = funcref.func

        # self.name = funcref.name
        self.local_namespace = get_function_context_dict(func).copy()

        paramspec = inspect.getfullargspec(funcref.func)
        args = dict(zip(paramspec.args, args))
        args.update(kwds)

        self.ret_dtype = None
        self.const_args = {}

        kwddefaults = paramspec.kwonlydefaults or {}

        if paramspec.defaults:
            for name, val in zip(reversed(paramspec.args), reversed(paramspec.defaults)):
                if name in args:
                    continue

                args[name] = ir.ResExpr(val)

        for k, v in kwddefaults.items():
            if k not in args:
                args[k] = ir.ResExpr(v)

        if func.__annotations__:
            params = {**func.__annotations__, **kwddefaults}

            for a in paramspec.args:
                if a not in params:
                    params[a] = Any

            arg_types = {}
            for name in paramspec.args:
                if name not in args:
                    continue

                arg_types[name] = args[name].dtype

            for name, var in args.items():
                if name in params:
                    continue

                if isinstance(var, ir.ResExpr):
                    params[name] = var.val
                else:
                    params[name] = var

            for n, v in kwds.items():
                if isinstance(v, ir.ResExpr):
                    params[n] = v.val

            res = infer_ftypes(params=params, args=arg_types, namespace=self.local_namespace)

            if 'return' in res:
                self.ret_dtype = res['return']

            for name, dtype in res.items():
                if name not in args:
                    continue

                if isinstance(args[name], ir.ResExpr):
                    self.local_namespace[name] = args[name].val
                    self.const_args[name] = args[name]
                else:
                    self.scope[name] = ir.Variable(name, dtype)
        else:
            for name, arg in args.items():
                if isinstance(arg, ir.ResExpr):
                    self.local_namespace[name] = arg.val
                    self.const_args[name] = arg
                else:
                    self.scope[name] = ir.Variable(name, arg.dtype)

        # self.scope maybe contains infered type, so use that if present
        self.signature = {
            name: self.scope.get(name, var).dtype
            for name, var in args.items() if name not in self.const_args
        }


def reraise(tp, value, tb=None):
    if value.__traceback__ is not tb:
        raise value.with_traceback(tb)
    raise value


def node_visitor(ast_type):
    def wrapper(f):
        def func_wrapper(node, ctx):
            if reg['trace/level'] == 0:
                return f(node, ctx)

            err = None

            try:
                return f(node, ctx)
            except Exception as e:
                if isinstance(e, HLSSyntaxError) and e.lineno is not None:
                    _, exc_value, tb = sys.exc_info()
                elif hasattr(node, 'lineno'):
                    exc_value, tb = form_hls_syntax_error(ctx, e, node.lineno)
                else:
                    err = e
                    tb = None

            if err is not None:
                raise err
            else:
                raise exc_value.with_traceback(tb)

        if isinstance(ast_type, tuple):
            f_ret = visit_ast.register(ast_type[0])(func_wrapper)
            for t in ast_type[1:]:
                f_ret = visit_ast.register(t)(f_ret)
        else:
            f_ret = visit_ast.register(ast_type)(func_wrapper)

        return f_ret

    return wrapper


@singledispatch
def visit_ast(node, ctx):
    """Used by default. Called if no explicit function exists for a node."""
    if node is None:
        return ir.ResExpr(None)

    breakpoint()
    raise SyntaxError(f"Unsupported language construct", node.lineno)


def visit_block(ir_node, body, ctx):
    ctx.new_closure(ir_node)
    for stmt in body:
        res_stmt = visit_ast(stmt, ctx)
        if not isinstance(res_stmt, list):
            res_stmt = [res_stmt]

        for s in res_stmt:
            if s is None:
                continue

            ir_node.stmts.append(s)
            if isinstance(ir_node, ir.FuncBlock) and isinstance(s, ir.FuncReturn):
                # No need to continue, return has been hit
                break

        if ir_node.stmts and isinstance(ir_node, ir.FuncBlock) and isinstance(
                ir_node.stmts[-1], ir.FuncReturn):
            # No need to continue, return has been hit
            break

    # Remove expressions that are added as block statements
    stmts = ir_node.stmts
    ir_node.stmts = []
    for s in stmts:
        if isinstance(s, ir.CallExpr):
            ir_node.stmts.append(ir.ExprStatement(s))
        elif isinstance(s, ir.Expr):
            pass
        else:
            ir_node.stmts.append(s)

    ctx.closures.pop()

    return ir_node



# from pygears.conf import safe_bind
from pygears.typing.base import typeof, is_type
from . import Array, Int, Integer, Queue, Tuple, Uint, Union, Maybe, Any
from . import Fixpnumber, Float, Number, Ufixp, Fixp, Unit, Integral
# from pygears.conf.log import gear_log

# TODO: Find solution for when a field of a complex data type needs to be
# recoded (for an example to a smaller width). This doesn't work properly


def short_s(s, cutoff=20):
    if len(s) > cutoff:
        return s[:(cutoff // 2 - 1)] + '...' + s[-(cutoff // 2 - 2):]

    return s


def short_repr(dtype, cutoff=20):
    return short_s(repr(dtype))


def get_type_error(dtype, cast_type, details=None):
    if details is None:
        details = []
    else:
        details = list(details)

    if getattr(cast_type, 'specified', False):
        details.append(
            f"FIX: to interpret '{short_repr(dtype)}' encoded value as '{short_repr(cast_type)}',"
            f" use code()")
    else:
        details.append(
            f"FIX: to interpret '{short_repr(dtype)}' encoded value as '{short_repr(cast_type)}',"
            f" specify the cast type completely and use code()")

    return TypeError(f"Cannot convert type '{repr(dtype)}' to '{repr(cast_type)}'\n    " +
                     '\n    '.join(details))


def get_value_error(val, cast_type):
    return ValueError(f"Cannot convert value '{repr(val)}' to '{repr(cast_type)}'")


def uint_type_cast_resolver(dtype, cast_type):
    if typeof(dtype, Ufixp):
        if cast_type.specified:
            if cast_type.width < dtype.integer:
                raise get_type_error(dtype, cast_type, [
                    f"fixed-point integer part (width '{dtype.integer}') is larger than target Uint",
                    f"FIX: to force cast to smaller Uint, cast to generic Uint first and then code() as '{repr(cast_type)}'"
                ])

            return cast_type
        elif dtype.integer <= 0:
            raise get_type_error(dtype, cast_type, [
                f"fixed-point has no integer part",
                f"FIX: to force cast to Uint, supply Uint width explicitly"
            ])
        else:
            return Uint[dtype.integer]

    if typeof(dtype, Uint):
        if not cast_type.specified:
            return dtype
        elif dtype.width > cast_type.width:
            raise get_type_error(dtype, cast_type,
                                 [f"{repr(dtype)} is larger then {repr(cast_type)}"])
        else:
            return cast_type

    raise get_type_error(dtype, cast_type)


def float_type_cast_resolver(dtype, cast_type):
    if typeof(dtype, Number):
        return cast_type

    raise get_type_error(dtype, cast_type)


def ufixp_type_cast_resolver(dtype, cast_type):
    if typeof(dtype, Ufixp):
        if not cast_type.specified:
            return dtype

        if dtype.integer > cast_type.integer:
            raise get_type_error(dtype, cast_type, [
                f"fixed-point integer part is larger than target's integer part ('{dtype.integer}' > '{cast_type.integer})",
            ])

        if dtype.fract > cast_type.fract:
            raise get_type_error(dtype, cast_type, [
                f"fixed-point fraction part is larger than target's fraction part ('{dtype.fract}' > '{cast_type.fract})",
            ])

        return cast_type

    if typeof(dtype, Uint):
        if not cast_type.specified:
            return Ufixp[dtype.width, dtype.width]

        if dtype.width > cast_type.integer:
            raise get_type_error(dtype, cast_type, [
                f"integer is larger than fixed-point integer part ('{dtype.width}' > '{cast_type.integer})",
            ])

        return cast_type

    raise get_type_error(dtype, cast_type)


def fixp_type_cast_resolver(dtype, cast_type):
    if typeof(dtype, Fixpnumber):
        if not cast_type.specified:
            if dtype.signed:
                return dtype
            else:
                return Fixp[dtype.integer + 1, dtype.width + 1]

        int_part = dtype.integer
        if not dtype.signed:
            int_part += 1

        if int_part > cast_type.integer:
            raise get_type_error(dtype, cast_type, [
                f"needed target's integer part >= {int_part}",
            ])

        if dtype.fract > cast_type.fract:
            raise get_type_error(dtype, cast_type, [
                f"fixed-point fraction part is larger than target's fraction part ('{dtype.fract}' > '{cast_type.fract})",
            ])

        return cast_type

    if typeof(dtype, Integer):
        dtype = type_cast(dtype, Int)

        if not cast_type.specified:
            return Fixp[dtype.width, dtype.width]

        if dtype.width > cast_type.integer:
            raise get_type_error(dtype, cast_type, [
                f"integer is larger than fixed-point integer part ('{dtype.width}' > '{cast_type.integer})",
            ])

        return cast_type

    raise get_type_error(dtype, cast_type)


def fixpnumber_type_cast_resolver(dtype, cast_type):
    if not typeof(dtype, Integral):
        raise get_type_error(dtype, cast_type)

    if dtype.signed:
        return fixp_type_cast_resolver(dtype, Fixp[cast_type.__args__])
    else:
        return ufixp_type_cast_resolver(dtype, Ufixp[cast_type.__args__])


def int_type_cast_resolver(dtype, cast_type):
    if typeof(dtype, Fixpnumber):
        int_part = dtype.integer
        if not dtype.signed:
            int_part += 1

        if cast_type.specified:
            if not dtype.signed and cast_type.width == dtype.integer:
                raise get_type_error(dtype, cast_type, (
                    f"Int needs to be one bit larger (width {int_part}) to represent unsigned fixed-point integer part (width {dtype.integer})",
                    f"FIX: to force cast to smaller Int, cast to generic Int first and then code() as '{repr(cast_type)}'"
                ))

            elif cast_type.width < dtype.integer:
                raise get_type_error(dtype, cast_type, (
                    f"fixed-point integer part (width '{dtype.integer}') is larger than target Int",
                    f"FIX: to force cast to smaller Int, cast to generic Int first and then code() as '{repr(cast_type)}'"
                ))

            return cast_type
        elif dtype.integer <= 0:
            raise get_type_error(dtype, cast_type,
                                 (f"fixed-point has no integer part",
                                  f"FIX: to force cast to Int, supply Int width explicitly"))
        else:
            return Int[int_part]

    if typeof(dtype, Integer):
        width = dtype.width
        if not dtype.signed:
            width += 1

        if not cast_type.specified:
            return Int[width]
        elif not dtype.signed and cast_type.width == dtype.width:
            raise get_type_error(dtype, cast_type, (
                f"Int needs to be one bit larger (width {width}) to represent unsigned integer (width {dtype.width})",
                f"FIX: to force cast to smaller Int, cast to generic Int first and then code() as '{repr(cast_type)}'"
            ))
        elif cast_type.width < dtype.width:
            raise get_type_error(dtype, cast_type,
                                 [f"{repr(dtype)} is larger then {repr(cast_type)}"])
        else:
            return cast_type

    raise get_type_error(dtype, cast_type)


def plain_int_type_cast_resolver(dtype, cast_type):
    if typeof(dtype, (Uint, Int)):
        return int

    raise get_type_error(dtype, cast_type)


def tuple_type_cast_resolver(dtype, cast_type):
    if typeof(dtype, Tuple) and len(cast_type) == 0:
        return dtype

    if typeof(dtype, (Queue, Union, Tuple)):
        fields = [d for d in dtype]
    elif typeof(dtype, Array):
        fields = [dtype.data] * len(dtype)
    else:
        raise get_type_error(dtype, cast_type)

    if len(cast_type) == 0:
        return Tuple[tuple(fields)]
    elif len(cast_type) != len(fields):
        comp = 'less' if len(cast_type) < len(fields) else 'more'
        raise get_type_error(
            dtype, cast_type,
            [f"target Tuple has {comp} elements ({len(cast_type)}) than needed ({len(fields)})"])
    else:
        try:
            cast_fields = [cast(dt, ct) for dt, ct in zip(fields, cast_type.args)]
        except TypeError as e:
            raise TypeError(f"{str(e)}\n    - when casting '{repr(dtype)}' to '{repr(cast_type)}'")

        if typeof(cast_type, Tuple):
            return Tuple[dict(zip(cast_type.fields, cast_fields))]
        else:
            return Tuple[tuple(cast_fields)]

    raise get_type_error(dtype, cast_type)


def array_type_cast_resolver(dtype, cast_type):
    if (typeof(dtype, Array) and (len(cast_type.args) == 0)):
        return dtype

    if typeof(dtype, Tuple):
        if len(cast_type.args) >= 1:
            arr_dtype = cast(dtype.args[0], cast_type.dtype)
        else:
            arr_dtype = dtype.args[0]

        if len(cast_type.args) == 2:
            if len(dtype) != len(cast_type):
                comp = 'less' if len(cast_type) < len(dtype) else 'more'
                raise get_type_error(dtype, cast_type, [
                    f"target Array has {comp} elements ({len(cast_type)}) than Tuple ({len(dtype)})"
                ])

        try:
            for t in dtype.args:
                cast(t, arr_dtype)
        except TypeError as e:
            raise TypeError(f"{str(e)}\n    - when casting '{repr(dtype)}' to '{repr(cast_type)}'")

        return Array[arr_dtype, len(dtype)]

    if typeof(dtype, Array):
        dlen, clen = len(dtype), len(cast_type)
        if dlen != clen:
            comp = 'less' if clen < dlen else 'more'
            err_detail = [
                f"target Array has {comp} elements ({len(cast_type)}) than source ({len(dtype)})"
            ]

            if dlen % clen == 0:
                subarray = Array[dtype.data, dlen // clen]
                try:
                    subarray = cast(subarray, cast_type.data)
                    return Array[subarray, clen]
                except TypeError as e:
                    err_detail.append(str(e))

            raise get_type_error(dtype, cast_type, err_detail)

        try:
            arr_dtype = cast(dtype.data, cast_type.data)
        except TypeError as e:
            raise TypeError(f"{str(e)}\n    - when casting '{repr(dtype)}' to '{repr(cast_type)}'")

        return Array[arr_dtype, len(dtype)]

    raise get_type_error(dtype, cast_type)


def union_type_cast_resolver(dtype, cast_type):
    if (typeof(dtype, Union) and (len(cast_type.args) == 0)):
        return dtype

    if typeof(dtype, Tuple):
        if len(dtype) != 2:
            raise get_type_error(dtype, cast_type,
                                 [f"only Tuple with exactly 2 elements can be converted to Union"])

        if typeof(cast_type, Maybe):
            if dtype.args[1].width != 1:
                raise get_type_error(
                    dtype, cast_type,
                    [f"only Tuple with 1 bit wide second argument can be converted to Maybe"])

            if not cast_type.specified:
                return Maybe[dtype.args[0]]

            data_type = cast(dtype.args[0], cast_type.types[1])

            return Maybe[data_type]

        if len(cast_type.types) != 0:
            cast_ctrl = cast_type.ctrl
            types = tuple(cast_type.types)

            if cast_type.data.width < dtype.args[0].width:
                raise get_type_error(dtype, cast_type,
                                     [f"Tuple first element larger than target Union data field"])

            try:
                ctrl = cast(dtype.args[1], cast_ctrl)
            except TypeError as e:
                raise TypeError(
                    f"{str(e)}\n    - when casting '{repr(dtype)}' to '{repr(cast_type)}'")
        else:
            ctrl = cast(dtype.args[1], Uint)

            if ctrl.width > 6:
                raise TypeError(f'Casting to large Union with {2**ctrl.width}'
                                f' subtypes from {dtype}')

            types = (dtype.args[0], ) * (2**ctrl.width)

        res = Union[types]

        return res

    if typeof(cast_type, Maybe):
        if not typeof(dtype, (Tuple, Union)):
            raise get_type_error(
                dtype, cast_type,
                [f"only Tuples and Unions can be converted to Maybe"])

        if not (len(dtype.types) == 2 and typeof(dtype.types[0], Unit)):
            raise get_type_error(
                dtype, cast_type,
                [f"only Union's with with the form 'Union[Unit, Any]' can be converted to Maybe"])

        return Maybe[dtype.types[1]]

    raise get_type_error(dtype, cast_type)


def queue_type_cast_resolver(dtype, cast_type):
    if typeof(dtype, Queue):
        if len(cast_type.args) == 0:
            return dtype

        if dtype.lvl != cast_type.lvl:
            raise get_type_error(
                dtype, cast_type,
                [f"Queue level ({dtype.lvl}) must match the cast Queue lelve ({cast_type.lvl})"])

        try:
            return Queue[cast(dtype.data, cast_type.data), cast_type.lvl]
        except TypeError as e:
            raise TypeError(f"{str(e)}\n    - when casting '{repr(dtype)}' to '{repr(cast_type)}'")

    if typeof(dtype, Tuple):
        if len(dtype) != 2:
            raise get_type_error(dtype, cast_type,
                                 [f"only Tuple with exactly 2 elements can be converted to Queue"])

        lvl = cast(dtype[1], Uint).width
        if len(cast_type.args) != 0:
            if cast_type.lvl != lvl:
                raise get_type_error(dtype, cast_type, [
                    f"second Tuple element width ({lvl}) must match Queue level ({cast_type.lvl})"
                ])

            try:
                data = cast(dtype.args[0], cast_type.data)
            except TypeError as e:
                raise TypeError(
                    f"{str(e)}\n    - when casting '{repr(dtype)}' to '{repr(cast_type)}'")
        else:
            data = dtype.args[0]

        return Queue[data, lvl]

    raise get_type_error(dtype, cast_type)


type_cast_resolvers = {
    Uint: uint_type_cast_resolver,
    Int: int_type_cast_resolver,
    Ufixp: ufixp_type_cast_resolver,
    Fixp: fixp_type_cast_resolver,
    Fixpnumber: fixpnumber_type_cast_resolver,
    Tuple: tuple_type_cast_resolver,
    Union: union_type_cast_resolver,
    Float: float_type_cast_resolver,
    float: float_type_cast_resolver,
    Array: array_type_cast_resolver,
    Queue: queue_type_cast_resolver,
    int: plain_int_type_cast_resolver,
    Any: lambda dtype, cast_type: dtype
}


def type_cast(dtype, cast_type):

    if dtype == cast_type:
        return dtype

    if isinstance(cast_type, str):
        return dtype

    for templ in type_cast_resolvers:
        if typeof(cast_type, templ):
            return type_cast_resolvers[templ](dtype, cast_type)

    raise TypeError(f"Cannot cast '{repr(dtype)}' to '{repr(cast_type)}'")


def integer_value_cast_resolver(val, cast_type):
    if isinstance(val, int):
        return cast_type(val)

    if typeof(type(val), Integer):
        if not cast_type.specified:
            cast_type = cast_type[val.width]

        tout_range = (1 << int(cast_type))
        val = int(val) & (tout_range - 1)

        if typeof(cast_type, Int):
            max_uint = tout_range / 2 - 1
            if val > max_uint:
                val -= tout_range

        return cast_type(val)

    raise get_value_error(val, cast_type)


def uint_value_cast_resolver(val, cast_type):
    val_type = type(val)
    if not is_type(val_type) and isinstance(val, (int, float)):
        return cast_type(int(val))

    cast_type = uint_type_cast_resolver(val_type, cast_type)

    if typeof(val_type, Ufixp):
        if val_type.fract >= 0:
            return cast_type.decode(val.code() >> val_type.fract)
        else:
            return cast_type.decode(val.code() << (-val_type.fract))

    if typeof(val_type, Uint):
        return cast_type(val)

    raise get_value_error(val, cast_type)


def int_value_cast_resolver(val, cast_type):
    val_type = type(val)
    if not is_type(val_type) and isinstance(val, (int, float)):
        return cast_type(int(val))

    cast_type = int_type_cast_resolver(val_type, cast_type)

    if typeof(val_type, Fixpnumber):
        if val_type.fract >= 0:
            return cast_type.decode(val.code() >> val_type.fract)
        else:
            return cast_type.decode(val.code() << (-val_type.fract))

    if typeof(val_type, Integer):
        return cast_type(val)

    raise get_value_error(val, cast_type)


def plain_int_value_cast_resolver(val, cast_type):
    return int(val)


def tuple_value_cast_resolver(val, cast_type):
    if (isinstance(val, (list, tuple, dict)) and not is_type(type(val))
            and not cast_type.specified):
        return cast_type(val)

    val_type = type(val)
    cast_type = tuple_type_cast_resolver(val_type, cast_type)

    return cast_type(tuple(cast(v, ct) for v, ct in zip(val, cast_type)))


def array_value_cast_resolver(val, cast_type):
    val_type = type(val)
    cast_type = array_type_cast_resolver(val_type, cast_type)
    cast_elem = cast_type.data

    dlen, clen = len(val_type), len(cast_type)
    if dlen == clen:
        return cast_type(tuple(cast(v, cast_elem) for v in val))
    else:
        elen = len(cast_elem)
        return cast_type(tuple(cast(val[elen*i:elen*(i+1)], cast_elem) for i in range(clen)))


def union_value_cast_resolver(val, cast_type):
    val_type = type(val)
    cast_type = union_type_cast_resolver(val_type, cast_type)

    data = cast_type.data(val[0].code())
    ctrl = cast_type.ctrl(val[1])
    return cast_type((data, ctrl))


def queue_value_cast_resolver(val, cast_type):
    val_type = type(val)
    cast_type = queue_type_cast_resolver(val_type, cast_type)

    return cast_type(tuple(cast(v, t) for v, t in zip(val, cast_type)))


def fixpnumber_value_cast_resolver(val, cast_type):
    val_type = type(val)
    cast_type = fixpnumber_type_cast_resolver(val_type, cast_type)

    return cast_type(val)


def float_value_cast_resolver(val, cast_type):
    if typeof(type(val), (int, float, Number)):
        return cast_type(val)

    raise Exception(f'Only numbers can be converted to Float, not {val} of the type'
                    f' {repr(type(val))}')


value_cast_resolvers = {
    Uint: uint_value_cast_resolver,
    Int: int_value_cast_resolver,
    Integer: integer_value_cast_resolver,
    Tuple: tuple_value_cast_resolver,
    Array: array_value_cast_resolver,
    Union: union_value_cast_resolver,
    Float: float_value_cast_resolver,
    float: float_value_cast_resolver,
    Fixpnumber: fixpnumber_value_cast_resolver,
    Queue: queue_value_cast_resolver,
    int: plain_int_value_cast_resolver,
    Any: lambda val, cast_type: val
}


def value_cast(val, cast_type):
    if type(val) == cast_type:
        return val

    for templ in value_cast_resolvers:
        if typeof(cast_type, templ):
            return value_cast_resolvers[templ](val, cast_type)

    raise ValueError(f"Type '{repr(cast_type)}' unsupported, cannot cast value '{val}' "
                     f"of type '{repr(type(val))}'")


def cast(data, cast_type):
    if is_type(data):
        return type_cast(data, cast_type)
    else:
        return value_cast(data, cast_type)


def signed(dtype_or_val):
    if dtype_or_val.signed:
        return dtype_or_val

    if is_type(dtype_or_val):
        if typeof(dtype_or_val, Uint):
            return cast(dtype_or_val, Int)


# class CastCoreTypesPlugin(CoreTypesPlugin):
#     @classmethod
#     def bind(cls):
#         reg['gear/type_arith/cast'] = cast



from pygears import module
from pygears.sim import delta
from pygears.typing import Queue, Tuple, typeof
from pygears import gear, alternative
from pygears.lib.shred import shred
from .ccat import ccat
from .permute import permuted_apply
from .cat_util import din_data_cat_value
from functools import reduce
from pygears.util.utils import gather


def lvl_if_queue(t):
    if not issubclass(t, Queue):
        return 0
    else:
        return t.lvl


def data_if_queue(t):
    if not issubclass(t, Queue):
        return t
    else:
        return t[0]


def zip_type(dtypes):
    arg_queue_lvl = list(map(lvl_if_queue, dtypes))
    base_type = Tuple[tuple(map(data_if_queue, dtypes))]

    # If there are no Queues, i.e. max(arg_queue_lvl) == 0, the type below
    # will resolve to just base_type
    return Queue[base_type, max(arg_queue_lvl)]


@gear
async def zip_cat(*din) -> b'zip_type(din)':
    id_max_lvl, max_lvl = max(enumerate(din),
                              key=lambda p: p[1].dtype.lvl if typeof(p[1].dtype, Queue) else 0)

    async with gather(*din) as dout:
        yield (din_data_cat_value(dout), dout[id_max_lvl].eot)


def isort(iterable, key=lambda x: x, reverse=False):
    res = sorted(enumerate(iterable), key=key, reverse=reverse)
    values = tuple(d[1] for d in res)
    indices = tuple(d[0] for d in res)

    return values, indices


@gear
def czip2(a, b) -> b'zip_type((a, b))':
    return (a, b) | zip_sync(outsync=False) | zip_cat


@gear
def czip(*din):
    if len(din) == 2:
        return czip2(*din)

    # Sort input interfaces in descending order of their Queue levels, i.e. we
    # want to zip highest Queue levels first in order to synchronize them first
    din_sorted_by_lvl, din_sort_indices = isort(din,
                                                key=lambda x: lvl_if_queue(x[1].dtype),
                                                reverse=True)

    # Zip din's in sorted order using it as a binary operation. This will
    # produce nested Tuple's, hence we cast it to a Queue of single Tuple
    ret_flat_type = zip_type([d.dtype for d in din_sorted_by_lvl])

    def czip_cascade(*din):
        return reduce(czip, din) >> ret_flat_type

    return permuted_apply(*din, f=czip_cascade, indices=din_sort_indices)


@gear
def unzip(din, *, dtypes):
    zdata, zlast = din

    def split():
        for i, d in enumerate(dtypes):
            data = zdata[i]
            if issubclass(d, Queue):
                yield ccat(data, zlast[:d.lvl]) | Queue[data.dtype, d.lvl]
            else:
                yield data

    return tuple(split())


@gear(enablement=b'len(din) == 2')
async def zip_sync(*din, outsync=True) -> b'din':
    lvls = tuple(d.dtype.lvl if typeof(d.dtype, Queue) else 0 for d in din)
    overlap_lvl = min(lvls)

    eot_aligned = (1, 1)

    while (1):
        din_data = [(await d.pull()) for d in din]

        if overlap_lvl > 0:
            eot_overlap = [d.eot[:overlap_lvl] for d in din_data]

            eot_aligned = (eot_overlap[0] >= eot_overlap[1], eot_overlap[1] >= eot_overlap[0])
        else:
            eot_aligned = (1, 1)
            eot_overlap = din_data[0].eot if lvls[0] else din_data[1].eot

        if all(eot_aligned):
            yield din_data
        else:
            await delta()

        for d, aligned in zip(din, eot_aligned):
            if (not aligned) or all(eot_aligned):
                d.ack()


@alternative(zip_sync)
@gear(enablement=b'len(din) > 2')
def zip_sync_vararg(*din):
    return din | czip | unzip(dtypes=[d.dtype for d in din])


@gear
def zip_sync_with(sync_in, din, *, balance=None):
    if balance:
        sync_in = sync_in | balance

    din_sync, sync_in_sync = zip_sync(din, sync_in)
    sync_in_sync | shred

    return din_sync


@gear
def zip_wrap_with(sync, din):
    din_zip = czip(sync, din)

    return ccat(din_zip['data'][1], din_zip['eot']) | Queue



from math import floor, ceil


def bitw(num: int) -> int:
    num = int(num)
    if num < 0:
        num = -2 * num - 1
    elif num == 0:
        return 1

    return num.bit_length()


def ceil_pow2(num: int) -> int:
    return int(2**(bitw(ceil(num) - 1)))


def ceil_div(num: int, divisor: int) -> int:
    return (num + divisor - 1) // divisor


def ceil_chunk(num: int, chunk: int) -> int:
    return ceil_div(num, chunk)*chunk


def div(a, b, subprec):
    res = a.__truediv__(b, subprec)
    if res is not NotImplemented:
        return res

    return b.__rtruediv__(a, subprec)



import logging
import platform
import sys
import textwrap
import inspect
from functools import partial
from traceback import TracebackException

from .log import CustomLogger, LogPlugin, register_custom_log
from .registry import Inject, inject, reg

from .trace_format import enum_traceback, TraceLevel, enum_stacktrace


def gear_definition_location(func):
    uwrp = inspect.unwrap(func, stop=(lambda f: hasattr(f, "__signature__")))
    fn = inspect.getfile(uwrp)

    ln = 1
    lines = None
    while fn.startswith('<decorator'):
        if hasattr(uwrp, 'definition'):
            uwrp = uwrp.definition
        elif hasattr(uwrp, 'alternative_to'):
            uwrp = uwrp.alternative_to
        else:
            break

        fn = inspect.getfile(uwrp)

    try:
        lines, ln = inspect.getsourcelines(uwrp)
    except OSError:
        pass

    return uwrp, fn, ln, lines


class MultiAlternativeError(Exception):
    def __init__(self, errors):
        self.errors = errors

    def __str__(self):
        ret = ['\n']
        for err_pack in self.errors:
            if err_pack is None:
                continue

            func, err_cls, err, tr = err_pack

            ret.append('\n')
            tr_list = None if not tr else list(enum_traceback(tr))
            if tr_list:
                ret.extend(tr_list)
            else:
                funcdef, fn, ln, _ = gear_definition_location(func)

                ret.append(
                    f'  File "{fn}", line {ln}, in {funcdef.__name__}\n')

            exc_msg = register_issue(err_cls, err)
            ret.append(textwrap.indent(exc_msg, 4 * ' '))

        return textwrap.indent(str(''.join(ret)), 2 * ' ')


@inject
def register_issue(err_cls, err, issues=Inject('trace/issues')):
    tr_exc = TracebackException(err_cls, err, None)
    issue_id = len(issues)
    if isinstance(err, MultiAlternativeError):
        issue = f"{err_cls.__name__}: {tr_exc}\n"
    else:
        issue = f"{err_cls.__name__}: [{issue_id}], {tr_exc}\n"
        issues.append(err)

    return issue


def register_exit_hook(hook, *args, **kwds):
    reg['trace/hooks'].append(partial(hook, *args, **kwds))


def log_exception(exception):
    exception_type = type(exception)
    tr = exception.__traceback__

    from pygears.conf.log import LogException
    print_traceback = (exception_type is not LogException)
    if not print_traceback:
        print_traceback = reg[f'logger/{exception.name}/print_traceback']
    if print_traceback:
        for s in enum_traceback(tr):
            logging.getLogger('trace').error(s[:-1])

    logging.getLogger('trace').error(register_issue(exception_type, exception))


def pygears_excepthook(exception_type,
                       exception,
                       tr,
                       debug_hook=sys.excepthook):

    for hook in reg['trace/hooks']:
        try:
            hook()
        except Exception:
            pass

    if reg['trace/level'] == TraceLevel.debug:
        debug_hook(exception_type, exception, tr)
    else:
        from pygears.util.print_hier import print_hier
        from pygears import find

        try:
            print_hier(find('/'))
        except Exception:
            pass

        log_exception(exception.with_traceback(tr))


@inject
def stack_trace(name,
                verbosity,
                message,
                stack_traceback_fn=Inject('logger/stack_traceback_fn')):
    with open(stack_traceback_fn, 'a') as f:
        delim = '-' * 50 + '\n'
        f.write(delim)
        f.write(f'{name} [{verbosity.upper()}] {message}\n\n')
        tr = ''
        for s in enum_stacktrace():
            tr += s
        f.write(tr)
        f.write(delim)


def log_error_to_file(name, severity, msg):
    if severity in ['error', 'warning']:
        stack_trace(name, severity, msg)


class TraceLog(CustomLogger):
    def get_format(self):
        return logging.Formatter('%(message)s')

    def get_filter(self):
        return None


class TracePlugin(LogPlugin):
    @classmethod
    def bind(cls):
        register_custom_log('trace', cls=TraceLog)
        reg['logger/hooks'].append(log_error_to_file)
        reg['trace/issues'] = []


# -*- coding: utf-8 -*-
"""
    jinja2.debug
    ~~~~~~~~~~~~

    Implements the debug interface for Jinja.  This module does some pretty
    ugly stuff with the Python traceback system in order to achieve tracebacks
    with correct line numbers, locals and contents.

    :copyright: (c) 2017 by the Jinja Team.
    :license: BSD, see LICENSE for more details.
"""
import sys
import traceback
from types import TracebackType, CodeType

class TraceException(Exception):
    """Raised to tell the user that there is a problem with the template."""

    def __init__(self, message, lineno=None, name=None, filename=None):
        Exception.__init__(self, message)
        self.lineno = lineno
        self.name = name
        self.filename = filename
        self.source = None
        self.message = message

    def __reduce__(self):
        # https://bugs.python.org/issue1692335 Exceptions that take
        # multiple required arguments have problems with pickling.
        # Without this, raises TypeError: __init__() missing 1 required
        # positional argument: 'lineno'
        return self.__class__, (self.message, self.lineno, self.name, self.filename)

# on pypy we can take advantage of transparent proxies
try:
    from __pypy__ import tproxy
except ImportError:
    tproxy = None


# how does the raise helper look like?
try:
    exec("raise TypeError, 'foo'")
except SyntaxError:
    raise_helper = 'raise __jinja_exception__[1]'
except TypeError:
    raise_helper = 'raise __jinja_exception__[0], __jinja_exception__[1]'


class TracebackFrameProxy(object):
    """Proxies a traceback frame."""

    def __init__(self, tb):
        self.tb = tb
        self._tb_next = None

    @property
    def tb_next(self):
        return self._tb_next

    def set_next(self, next):
        if tb_set_next is not None:
            try:
                tb_set_next(self.tb, next and next.tb or None)
            except Exception:
                # this function can fail due to all the hackery it does
                # on various python implementations.  We just catch errors
                # down and ignore them if necessary.
                pass
        self._tb_next = next

    @property
    def is_jinja_frame(self):
        return '__jinja_template__' in self.tb.tb_frame.f_globals

    def __getattr__(self, name):
        return getattr(self.tb, name)


def make_frame_proxy(frame):
    proxy = TracebackFrameProxy(frame)
    if tproxy is None:
        return proxy
    def operation_handler(operation, *args, **kwargs):
        if operation in ('__getattribute__', '__getattr__'):
            return getattr(proxy, args[0])
        elif operation == '__setattr__':
            proxy.__setattr__(*args, **kwargs)
        else:
            return getattr(proxy, operation)(*args, **kwargs)
    return tproxy(TracebackType, operation_handler)


class ProcessedTraceback(object):
    """Holds a Jinja preprocessed traceback for printing or reraising."""

    def __init__(self, exc_type, exc_value, frames):
        assert frames, 'no frames for this traceback?'
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.frames = frames

        # newly concatenate the frames (which are proxies)
        prev_tb = None
        for tb in self.frames:
            if prev_tb is not None:
                prev_tb.set_next(tb)
            prev_tb = tb
        prev_tb.set_next(None)

    def render_as_text(self, limit=None):
        """Return a string with the traceback."""
        lines = traceback.format_exception(self.exc_type, self.exc_value,
                                           self.frames[0], limit=limit)
        return ''.join(lines).rstrip()

    def render_as_html(self, full=False):
        """Return a unicode string with the traceback as rendered HTML."""
        from jinja2.debugrenderer import render_traceback
        return u'%s\n\n<!--\n%s\n-->' % (
            render_traceback(self, full=full),
            self.render_as_text().decode('utf-8', 'replace')
        )

    @property
    def is_template_syntax_error(self):
        """`True` if this is a template syntax error."""
        return isinstance(self.exc_value, TraceException)

    @property
    def exc_info(self):
        """Exception info tuple with a proxy around the frame objects."""
        return self.exc_type, self.exc_value, self.frames[0]

    @property
    def standard_exc_info(self):
        """Standard python exc_info for re-raising"""
        tb = self.frames[0]
        # the frame will be an actual traceback (or transparent proxy) if
        # we are on pypy or a python implementation with support for tproxy
        if type(tb) is not TracebackType:
            tb = tb.tb
        return self.exc_type, self.exc_value, tb


def make_traceback(exc_info, source_hint=None):
    """Creates a processed traceback object from the exc_info."""
    exc_type, exc_value, tb = exc_info
    if isinstance(exc_value, TraceException):
        exc_info = translate_syntax_error(exc_value, source_hint)
        initial_skip = 0
    else:
        initial_skip = 1
    return translate_exception(exc_info, initial_skip)


def translate_syntax_error(error, source=None):
    """Rewrites a syntax error to please traceback systems."""
    error.source = source
    error.translated = True
    exc_info = (error.__class__, error, None)
    filename = error.filename
    if filename is None:
        filename = '<unknown>'
    return fake_exc_info(exc_info, filename, error.lineno)


def translate_exception(exc_info, initial_skip=0):
    """If passed an exc_info it will automatically rewrite the exceptions
    all the way down to the correct line numbers and frames.
    """
    tb = exc_info[2]
    frames = []

    # skip some internal frames if wanted
    for x in range(initial_skip):
        if tb is not None:
            tb = tb.tb_next
    initial_tb = tb

    while tb is not None:
        frames.append(make_frame_proxy(tb))
        tb = tb.tb_next

    return ProcessedTraceback(exc_info[0], exc_info[1], frames)


def fake_exc_info(exc_info, filename, lineno):
    """Helper for `translate_exception`."""
    exc_type, exc_value, tb = exc_info
    locals = {}

    # assamble fake globals we need
    globals = {
        '__name__':             filename,
        '__file__':             filename,
        '__jinja_exception__':  exc_info[:2],
    }

    # and fake the exception
    code = compile('\n' * (lineno - 1) + raise_helper, filename, 'exec')

    # if it's possible, change the name of the code.  This won't work
    # on some python environments such as google appengine
    try:
        if tb is None:
            location = 'template'
        else:
            function = tb.tb_frame.f_code.co_name
            if function == 'root':
                location = 'top-level template code'
            elif function.startswith('block_'):
                location = 'block "%s"' % function[6:]
            else:
                location = 'template'

        code = CodeType(0, code.co_kwonlyargcount,
                        code.co_nlocals, code.co_stacksize,
                        code.co_flags, code.co_code, code.co_consts,
                        code.co_names, code.co_varnames, filename,
                        location, code.co_firstlineno,
                        code.co_lnotab, (), ())
    except Exception as e:
        pass

    # execute the code and catch the new traceback
    try:
        exec(code, globals, locals)
    except:
        exc_info = sys.exc_info()
        new_tb = exc_info[2].tb_next

    # return without this frame
    return exc_info[:2] + (new_tb,)


if sys.version_info >= (3, 7):
    # tb_next is directly assignable as of Python 3.7
    def tb_set_next(tb, tb_next):
        tb.tb_next = tb_next
        return tb

elif platform.python_implementation() == "PyPy":
    # PyPy might have special support, and won't work with ctypes.
    try:
        import tputil
    except ImportError:
        # Without tproxy support, use the original traceback.
        def tb_set_next(tb, tb_next):
            return tb

    else:
        # With tproxy support, create a proxy around the traceback that
        # returns the new tb_next.
        def tb_set_next(tb, tb_next):
            def controller(op):
                if op.opname == "__getattribute__" and op.args[0] == "tb_next":
                    return tb_next

                return op.delegate()

            return tputil.make_proxy(controller, obj=tb)


else:
    # Use ctypes to assign tb_next at the C level since it's read-only
    # from Python.
    import ctypes

    class _CTraceback(ctypes.Structure):
        _fields_ = [
            # Extra PyObject slots when compiled with Py_TRACE_REFS.
            ("PyObject_HEAD", ctypes.c_byte * object().__sizeof__()),
            # Only care about tb_next as an object, not a traceback.
            ("tb_next", ctypes.py_object),
        ]

    def tb_set_next(tb, tb_next):
        c_tb = _CTraceback.from_address(id(tb))

        # Clear out the old tb_next.
        if tb.tb_next is not None:
            c_tb_next = ctypes.py_object(tb.tb_next)
            c_tb.tb_next = ctypes.py_object()
            ctypes.pythonapi.Py_DecRef(c_tb_next)

        # Assign the new tb_next.
        if tb_next is not None:
            c_tb_next = ctypes.py_object(tb_next)
            ctypes.pythonapi.Py_IncRef(c_tb_next)
            c_tb.tb_next = c_tb_next

        return tb



from pygears.hdl.util import HDLGearHierVisitor, flow_visitor
# from pygears.rtl import flow_visitor, RTLPlugin


@flow_visitor
class RemoveTupleFlattenVisitor(HDLGearHierVisitor):
    def flatten_tuple(self, node):
        node.bypass()


# class RTLFlattenPlugin(RTLPlugin):
#     @classmethod
#     def bind(cls):
#         cls.registry['rtl']['flow'].append(RemoveTupleFlattenVisitor)
